from dataclasses import dataclass
from enum import Enum


class SignatureType(Enum):
    ethereum_signature = "EthereumSignature"
    EIP1271_signature = "EIP1271Signature"


@dataclass
class TxEthSignature:
    sig_type: SignatureType
    signature: str

    @classmethod
    def from_dict(cls, json: dict):
        """
        Only the difference from __init__ that signature is already in hex format
        """
        obj = cls(sig_type=SignatureType(json["type"]), signature=b"")
        obj.signature = json["signature"]
        return obj

    def __init__(self, sig_type: SignatureType, signature: bytes):
        self.signature = signature.hex()
        self.sig_type = sig_type

    def dict(self):
        return {
            "type":      self.sig_type.value,
            "signature": self.signature
        }


@dataclass
class TxSignature:
    public_key: str
    signature: str

    @classmethod
    def from_dict(cls, json: dict):
        """
        Only the difference from __init__ is that values are already in hex format
        """
        obj = cls(public_key=b"", signature=b"")
        obj.public_key = json["pubKey"]
        obj.signature = json["signature"]
        return obj

    def __init__(self, public_key: bytes, signature: bytes):
        self.public_key = public_key.hex()
        self.signature = signature.hex()

    def dict(self):
        return {
            "pubKey":    self.public_key,
            "signature": self.signature
        }



from typing import Any, Dict, Optional
from enum import Enum
from decimal import Decimal
from zksync_sdk.types.transactions import Token

from pydantic import BaseModel


def to_camel(string: str) -> str:
    first, *others = string.split('_')
    return ''.join([first.lower(), *map(str.title, others)])


class Balance(BaseModel):
    amount: int
    expected_accept_block: int

    class Config:
        alias_generator = to_camel


class Depositing(BaseModel):
    balances: Dict[str, Balance]


class NFT(Token):
    creator_id: int
    content_hash: str
    creator_address: str
    serial_id: int
    decimals = 0

    def decimal_amount(self, amount: int) -> Decimal:
        return Decimal(amount)

    class Config:
        alias_generator = to_camel


class State(BaseModel):
    nonce: int
    pub_key_hash: str
    balances: Dict[str, int]
    nfts: Dict[str, NFT]
    minted_nfts: Dict[str, NFT]

    class Config:
        alias_generator = to_camel


class AccountTypes(str, Enum):
    OWNED = "Owned",
    CREATE2 = "CREATE2",
    NO_2FA = "No2FA"


class AccountState(BaseModel):
    address: str
    id: Optional[int]
    account_type: Optional[AccountTypes]
    depositing: Optional[Depositing]
    committed: Optional[State]
    verified: Optional[State]

    class Config:
        alias_generator = to_camel

    def get_nonce(self) -> int:
        assert self.committed is not None, "`get_nonce` needs `committed` to be set"
        return self.committed.nonce


class Fee(BaseModel):
    fee_type: Any
    gas_tx_amount: int
    gas_price_wei: int
    gas_fee: int
    zkp_fee: int
    total_fee: int

    class Config:
        alias_generator = to_camel


class ContractAddress(BaseModel):
    main_contract: str
    gov_contract: str

    class Config:
        alias_generator = to_camel


class BlockInfo(BaseModel):
    block_number: int
    committed: bool
    verified: bool

    class Config:
        alias_generator = to_camel


class EthOpInfo(BaseModel):
    executed: bool
    block: BlockInfo


class TransactionDetails(BaseModel):
    executed: bool
    success: bool
    fail_reason: Optional[str] = None
    block: BlockInfo

    class Config:
        alias_generator = to_camel



from .ethereum_provider import EthereumProvider
from .ethereum_signer import EthereumSignerWeb3
from .lib import ZkSyncLibrary
from .transport.http import HttpJsonRPCTransport
from .wallet import Wallet
from .zksync import ZkSync
from .zksync_provider import ZkSyncProviderV01
from .zksync_signer import ZkSyncSigner



from abc import ABC, abstractmethod
from typing import Any, List, Optional


class ProviderError(Exception):
    def __init__(self, basic_response, *args):
        self.basic_response = basic_response
        super().__init__(*args)


class ResponseError(Exception):
    def __init__(self, code, text, *args):
        self.code = code
        self.text = text
        super().__init__(*args)

    def __str__(self):
        return f"Response error with code {self.code} \n {self.text}"


class JsonRPCTransport(ABC):
    @abstractmethod
    async def request(self, method: str, params: Optional[List]) -> Any:
        pass

    def create_request(self, method: str, params=None):
        return {
            "id":      1,
            "jsonrpc": '2.0',
            "method":  method,
            "params":  params
        }



import ctypes
from ctypes import (Structure, c_ubyte, cdll)
import os
from typing import Optional

PRIVATE_KEY_LEN = 32
PUBLIC_KEY_LEN = 32
PUBKEY_HASH_LEN = 20
PACKED_SIGNATURE_LEN = 64
ORDER_LEN = 89
ORDERS_HASH_LEN = 31


class ZksPrivateKey(Structure):
    _fields_ = [
        ("data", c_ubyte * PRIVATE_KEY_LEN),
    ]


class ZksPackedPublicKey(Structure):
    _fields_ = [
        ("data", c_ubyte * PUBLIC_KEY_LEN),
    ]


class ZksPubkeyHash(Structure):
    _fields_ = [
        ("data", c_ubyte * PUBKEY_HASH_LEN),
    ]


class ZksSignature(Structure):
    _fields_ = [
        ("data", c_ubyte * PACKED_SIGNATURE_LEN),
    ]


class ZksOrdersHash(Structure):
    _fields_ = [
        ("data", c_ubyte * ORDERS_HASH_LEN),
    ]


class ZksOrders(Structure):
    _fields_ = [
        ("data", c_ubyte * (ORDER_LEN * 2)),
    ]


class ZkSyncLibrary:

    def __init__(self, library_path: Optional[str] = None):
        if library_path is None:
            library_path = os.environ["ZK_SYNC_LIBRARY_PATH"]
        self.lib = cdll.LoadLibrary(library_path)

    def private_key_from_seed(self, seed: bytes):
        private_key = ctypes.pointer(ZksPrivateKey())
        self.lib.zks_crypto_private_key_from_seed(seed, len(seed), private_key)
        return bytes(private_key.contents.data)

    def get_public_key(self, private_key: bytes):
        assert len(private_key) == PRIVATE_KEY_LEN
        public_key = ctypes.pointer(ZksPackedPublicKey())
        pk = ctypes.pointer(ZksPrivateKey(data=(c_ubyte * PRIVATE_KEY_LEN)(*private_key)))
        self.lib.zks_crypto_private_key_to_public_key(pk, public_key)
        return bytes(public_key.contents.data)

    def get_pubkey_hash(self, public_key: bytes):
        assert len(public_key) == PUBLIC_KEY_LEN
        public_key_hash = ctypes.pointer(ZksPubkeyHash())
        public_key_ptr = ctypes.pointer(
            ZksPackedPublicKey(data=(c_ubyte * PUBLIC_KEY_LEN)(*public_key)))
        self.lib.zks_crypto_public_key_to_pubkey_hash(public_key_ptr, public_key_hash)
        return bytes(public_key_hash.contents.data)

    def sign(self, private_key: bytes, message: bytes):
        assert len(private_key) == PRIVATE_KEY_LEN
        signature = ctypes.pointer(ZksSignature())
        private_key_ptr = ctypes.pointer(
            ZksPrivateKey(data=(c_ubyte * PRIVATE_KEY_LEN)(*private_key)))
        self.lib.zks_crypto_sign_musig(private_key_ptr, message, len(message), signature)
        return bytes(signature.contents.data)

    def hash_orders(self, orders: bytes):
        assert len(orders) == ORDER_LEN * 2
        orders_hash = ctypes.pointer(ZksOrdersHash())
        orders_bytes = ctypes.pointer(
            ZksOrders(data=(c_ubyte * (ORDER_LEN * 2))(*orders)))
        self.lib.rescue_hash_orders(orders_bytes, len(orders), orders_hash)
        return bytes(orders_hash.contents.data)

    def is_valid_signature(self, message: bytes, public_key: bytes, zk_sync_signature: bytes) -> bool:
        assert len(public_key) == PUBLIC_KEY_LEN
        assert len(zk_sync_signature) == PACKED_SIGNATURE_LEN
        public_key_ptr = ctypes.pointer(
            ZksPackedPublicKey(data=(c_ubyte * PUBLIC_KEY_LEN)(*public_key)))
        signature_ptr = ctypes.pointer(
            ZksSignature(data=(c_ubyte * PACKED_SIGNATURE_LEN)(*zk_sync_signature)))
        ret = self.lib.zks_crypto_verify_musig(message, len(message), public_key_ptr, signature_ptr)
        return ret == 0



from enum import Enum
from dataclasses import dataclass
from typing import Optional
from zksync_sdk.types.signatures import TxEthSignature


class ChangePubKeyTypes(Enum):
    onchain = "Onchain"
    ecdsa = "ECDSA"
    create2 = "CREATE2"


@dataclass
class ChangePubKeyEcdsa:
    batch_hash: bytes = b"\x00" * 32

    def encode_message(self) -> bytes:
        return self.batch_hash

    def dict(self, signature: str):
        return {"type": "ECDSA",
                "ethSignature": signature,
                "batchHash": f"0x{self.batch_hash.hex()}"}


@dataclass
class ChangePubKeyCREATE2:
    creator_address: str
    salt_arg: bytes
    code_hash: bytes

    def encode_message(self) -> bytes:
        return self.salt_arg

    def dict(self):
        return {"type": "CREATE2",
                "saltArg": f"0x{self.salt_arg.hex()}",
                "codeHash": f"0x{self.code_hash.hex()}"}


@dataclass
class Toggle2FA:
    enable: bool
    account_id: int
    time_stamp_milliseconds: int
    signature: TxEthSignature
    pub_key_hash: Optional[str]

    def dict(self):
        if self.pub_key_hash is not None:
            return {
                "enable": self.enable,
                "accountId": self.account_id,
                "timestamp": self.time_stamp_milliseconds,
                "signature": self.signature.dict(),
                "pubKeyHash": self.pub_key_hash
            }
        else:
            return {
                "enable": self.enable,
                "accountId": self.account_id,
                "timestamp": self.time_stamp_milliseconds,
                "signature": self.signature.dict(),
            }


def get_toggle_message(require_2fa: bool, time_stamp: int) -> str:
    if require_2fa:
        msg = f"By signing this message, you are opting into Two-factor Authentication protection by the zkSync " \
              f"Server.\n" \
              f"Transactions now require signatures by both your L1 and L2 private key.\n" \
              f"Timestamp: {time_stamp}"
    else:
        msg = f"You are opting out of Two-factor Authentication protection by the zkSync Server.\n" \
              f"Transactions now only require signatures by your L2 private key.\n" \
              f"BY SIGNING THIS MESSAGE, YOU ARE TRUSTING YOUR WALLET CLIENT TO KEEP YOUR L2 PRIVATE KEY SAFE!\n" \
              f"Timestamp: {time_stamp}"
    return msg


def get_toggle_message_with_pub(require_2fa: bool, time_stamp: int, pub_key_hash: str) -> str:
    msg = get_toggle_message(require_2fa, time_stamp)
    msg += f"\nPubKeyHash: {pub_key_hash}"
    return msg



from eth_account.messages import encode_defunct
from eth_account.signers.base import BaseAccount

from zksync_sdk import ZkSyncLibrary
from zksync_sdk.types import ChainId, EncodedTx, TxSignature


def derive_private_key(library: ZkSyncLibrary, message: str, account: BaseAccount,
                       chain_id: ChainId):
    if chain_id != ChainId.MAINNET:
        message = f"{message}\nChain ID: {chain_id}."
    signable_message = encode_defunct(message.encode())
    signature = account.sign_message(signable_message)
    private_key = library.private_key_from_seed(signature.signature)
    return private_key


class ZkSyncSigner:
    MESSAGE = "Access zkSync account.\n\nOnly sign this message for a trusted client!"

    def __init__(self, library: ZkSyncLibrary, private_key: bytes):
        self.library = library
        self.private_key = private_key
        self.public_key = self.library.get_public_key(self.private_key)

    @classmethod
    def from_account(cls, account: BaseAccount, library: ZkSyncLibrary, chain_id: ChainId):
        private_key = derive_private_key(library, cls.MESSAGE, account, chain_id)
        return cls(
            library=library,
            private_key=private_key,
        )

    @classmethod
    def from_seed(cls, library: ZkSyncLibrary, seed: bytes):
        private_key = library.private_key_from_seed(seed)
        return cls(
            library=library,
            private_key=private_key,
        )

    def pubkey_hash_str(self):
        return f"sync:{self.pubkey_hash().hex()}"

    def pubkey_hash(self):
        return self.library.get_pubkey_hash(self.public_key)

    def sign_tx(self, message: EncodedTx) -> TxSignature:
        signature = self.library.sign(self.private_key, message.encoded_message())
        return TxSignature(signature=signature, public_key=self.public_key)



from unittest import TestCase

from zksync_sdk.serializers import \
    num_to_bits, bits_into_bytes_in_be_order, reverse_bits, \
    closest_greater_or_eq_packable_amount, closest_greater_or_eq_packable_fee, closest_packable_transaction_fee


class TestNumberToBinaryArray(TestCase):

    def test_simple_conversion(self):
        bin_representation = num_to_bits(8, 4)
        self.assertListEqual(bin_representation, [0, 0, 0, 1])

        bin_representation = num_to_bits(32 + 5, 6)
        self.assertListEqual(bin_representation, [1, 0, 1, 0, 0, 1])

    def test_binary_list_to_bytes(self):
        # INFO: 32
        byte_array = bits_into_bytes_in_be_order([0, 0, 0, 1, 0, 0, 0, 0])
        self.assertEqual(byte_array[0], 0x10)

        values = [0, 0, 0, 0, 1, 0, 0, 0,
                  0, 0, 0, 1, 0, 0, 0, 0,
                  0, 0, 1, 0, 0, 0, 0, 0,
                  0, 1, 0, 0, 0, 0, 0, 0,
                  1, 0, 0, 0, 0, 0, 0, 0
                  ]
        byte_array = bits_into_bytes_in_be_order(values)
        self.assertEqual(len(byte_array), 5)
        self.assertEqual(byte_array[0], 8)
        self.assertEqual(byte_array[1], 16)
        self.assertEqual(byte_array[2], 32)
        self.assertEqual(byte_array[3], 64)
        self.assertEqual(byte_array[4], 128)

    def test_revers_bits(self):
        reverted = reverse_bits([1, 0, 0, 0, 1, 1, 0, 0])
        self.assertListEqual(reverted, [0, 0, 1, 1, 0, 0, 0, 1])

    def test_closest_greater_or_packable_all(self):
        nums = [0, 1, 2, 2047000, 1000000000000000000000000000000000]
        for num in nums:
            ret = closest_greater_or_eq_packable_amount(num)
            self.assertEqual(ret, num)
            ret = closest_greater_or_eq_packable_fee(num)
            self.assertEqual(ret, num)

    def test_closest_greater_or_packable_fee(self):
        ret = closest_greater_or_eq_packable_fee(2048)
        self.assertEqual(ret, 2050)
        ret = closest_packable_transaction_fee(2048)
        self.assertEqual(ret, 2047)



import importlib.resources as pkg_resources
import json

from . import contract_abi

zksync_abi_cache = None
ierc20_abi_cache = None

__all__ = ['zksync_abi', 'erc20_abi']


def zksync_abi():
    global zksync_abi_cache

    if zksync_abi_cache is None:
        abi_text = pkg_resources.read_text(contract_abi, 'ZkSync.json')
        zksync_abi_cache = json.loads(abi_text)['abi']

    return zksync_abi_cache


def erc20_abi():
    global ierc20_abi_cache

    if ierc20_abi_cache is None:
        abi_text = pkg_resources.read_text(contract_abi, 'IERC20.json')
        ierc20_abi_cache = json.loads(abi_text)['abi']

    return ierc20_abi_cache



from dataclasses import dataclass

from zksync_sdk.zksync_provider import FeeTxType
from zksync_sdk.wallet import Wallet, DEFAULT_VALID_FROM, DEFAULT_VALID_UNTIL, AmountsMissing
from zksync_sdk.types import (ChangePubKey, ChangePubKeyCREATE2, ChangePubKeyEcdsa,
                              ChangePubKeyTypes, EncodedTx, ForcedExit, TokenLike,
                              TransactionWithOptionalSignature,
                              Transfer, TxEthSignature,
                              Withdraw, MintNFT, WithdrawNFT, NFT, EncodedTxType, Order, Swap)
from typing import List, Union, Tuple, Optional
from decimal import Decimal


@dataclass
class BatchResult:
    transactions: list
    signature: TxEthSignature
    total_fees: dict


class BatchBuilder:
    IS_ENCODED_TRANSACTION = "is_encoded_trx"
    ENCODED_TRANSACTION_TYPE = "internal_type"

    TRANSACTIONS_ENTRY = "transactions"
    SIGNATURE_ENTRY = "signature"

    @classmethod
    def from_wallet(cls, wallet: Wallet, nonce: int, txs: Optional[List[EncodedTx]] = None):
        obj = BatchBuilder(wallet, nonce, txs)
        return obj

    def __init__(self, wallet: Wallet, nonce: int, txs: Optional[List[EncodedTx]] = None):
        if txs is None:
            txs = []
        self.wallet = wallet
        self.nonce = nonce
        self.batch_nonce = nonce
        self.transactions: List[dict] = []
        for tx in txs:
            value = tx.dict()
            value[self.IS_ENCODED_TRANSACTION] = True
            value[self.ENCODED_TRANSACTION_TYPE] = tx.tx_type()
            self.transactions.append(value)

    async def build(self) -> BatchResult:
        if not self.transactions:
            raise RuntimeError("Transaction batch cannot be empty")
        res = await self._process_transactions()
        trans = res["trans"]
        signature = self.wallet.eth_signer.sign(res["msg"].encode())
        return BatchResult(trans, signature, res["total_fee"])

    def add_withdraw(self,
                     eth_address: str,
                     token: TokenLike,
                     amount: Decimal,
                     fee: Optional[Decimal] = None,
                     valid_from=DEFAULT_VALID_FROM,
                     valid_until=DEFAULT_VALID_UNTIL
                     ):
        withdraw = {
            self.ENCODED_TRANSACTION_TYPE: EncodedTxType.WITHDRAW,
            self.IS_ENCODED_TRANSACTION: False,
            "eth_address": eth_address,
            "token": token,
            "amount": amount,
            "fee": fee,
            "valid_from": valid_from,
            "valid_until": valid_until
        }
        self.transactions.append(withdraw)

    def add_mint_nft(self,
                     content_hash: str,
                     recipient: str,
                     fee_token: TokenLike,
                     fee: Optional[Decimal] = None
                     ):
        mint_nft = {
            self.ENCODED_TRANSACTION_TYPE: EncodedTxType.MINT_NFT,
            self.IS_ENCODED_TRANSACTION: False,
            "content_hash": content_hash,
            "recipient": recipient,
            "fee_token": fee_token,
            "fee": fee
        }
        self.transactions.append(mint_nft)

    def add_withdraw_nft(self,
                         to_address: str,
                         nft_token: NFT,
                         fee_token: TokenLike,
                         fee: Optional[Decimal] = None,
                         valid_from=DEFAULT_VALID_FROM,
                         valid_until=DEFAULT_VALID_UNTIL
                         ):
        withdraw_nft = {
            self.ENCODED_TRANSACTION_TYPE: EncodedTxType.WITHDRAW_NFT,
            self.IS_ENCODED_TRANSACTION: False,
            "to_address": to_address,
            "nft_token": nft_token,
            "fee_token": fee_token,
            "fee": fee,
            "valid_from": valid_from,
            "valid_until": valid_until
        }
        self.transactions.append(withdraw_nft)

    def add_swap(self,
                 orders: Tuple[Order, Order],
                 fee_token: TokenLike,
                 amounts: Optional[Tuple[Decimal, Decimal]] = None,
                 fee: Optional[Decimal] = None
                 ):
        if amounts is None:
            if orders[0].amount == 0 or orders[1].amount == 0:
                raise AmountsMissing("in this case you must specify amounts explicitly")
        swap = {
            self.ENCODED_TRANSACTION_TYPE: EncodedTxType.SWAP,
            self.IS_ENCODED_TRANSACTION: False,
            "orders": orders,
            "fee_token": fee_token,
            "amounts": amounts,
            "fee": fee
        }
        self.transactions.append(swap)

    def add_transfer(self,
                     address_to: str,
                     token: TokenLike,
                     amount: Decimal,
                     fee: Optional[Decimal] = None,
                     valid_from=DEFAULT_VALID_FROM,
                     valid_until=DEFAULT_VALID_UNTIL
                     ):
        transfer = {
            self.ENCODED_TRANSACTION_TYPE: EncodedTxType.TRANSFER,
            self.IS_ENCODED_TRANSACTION: False,
            "from_address": self.wallet.address(),
            "to_address": address_to.lower(),
            "token": token,
            "amount": amount,
            "fee": fee,
            "valid_from": valid_from,
            "valid_until": valid_until
        }
        self.transactions.append(transfer)

    def add_change_pub_key(self,
                           fee_token: TokenLike,
                           eth_auth_type: Union[ChangePubKeyCREATE2, ChangePubKeyEcdsa, None],
                           fee: Optional[Decimal] = None,
                           valid_from=DEFAULT_VALID_FROM,
                           valid_until=DEFAULT_VALID_UNTIL
                           ):
        new_pubkey_hash = self.wallet.zk_signer.pubkey_hash_str()
        change_pub_key = {
            self.ENCODED_TRANSACTION_TYPE: EncodedTxType.CHANGE_PUB_KEY,
            self.IS_ENCODED_TRANSACTION: False,
            "account": self.wallet.address(),
            "new_pk_hash": new_pubkey_hash,
            "fee_token": fee_token,
            "fee": fee,
            "eth_auth_type": eth_auth_type,
            "valid_from": valid_from,
            "valid_until": valid_until
        }
        self.transactions.append(change_pub_key)

    def add_force_exit(self,
                       target_address: str,
                       token: TokenLike,
                       fee: Optional[Decimal] = None,
                       valid_from=DEFAULT_VALID_FROM,
                       valid_until=DEFAULT_VALID_UNTIL
                       ):
        forced_exit = {
            self.ENCODED_TRANSACTION_TYPE: EncodedTxType.FORCED_EXIT,
            self.IS_ENCODED_TRANSACTION: False,
            "target": target_address,
            "token": token,
            "fee": fee,
            "valid_from": valid_from,
            "valid_until": valid_until
        }
        self.transactions.append(forced_exit)

    async def _process_change_pub_key(self, obj: dict):
        if not obj[self.IS_ENCODED_TRANSACTION]:
            account_id = await self.wallet.get_account_id()
            token = await self.wallet.resolve_token(obj["fee_token"])
            eth_auth_type = obj["eth_auth_type"]
            if isinstance(eth_auth_type, ChangePubKeyEcdsa):
                eth_auth_type = ChangePubKeyTypes.ecdsa
            elif isinstance(eth_auth_type, ChangePubKeyCREATE2):
                eth_auth_type = ChangePubKeyTypes.create2
            else:
                eth_auth_type = ChangePubKeyTypes.onchain
            fee = obj["fee"]
            if fee is None:
                if eth_auth_type == ChangePubKeyTypes.ecdsa:
                    fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.change_pub_key_ecdsa,
                                                                            self.wallet.address(),
                                                                            token.id)
                elif eth_auth_type == ChangePubKeyTypes.onchain:
                    fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.change_pub_key_onchain,
                                                                            self.wallet.address(),
                                                                            token.id)
                elif eth_auth_type == ChangePubKeyTypes.create2:
                    fee = await self.wallet.zk_provider.get_transaction_fee(
                        FeeTxType.change_pub_key_create2,
                        self.wallet.address(),
                        token.id)
                fee = fee.total_fee
            else:
                fee = token.from_decimal(fee)
            change_pub_key = ChangePubKey(
                account=obj["account"],
                account_id=account_id,
                new_pk_hash=obj["new_pk_hash"],
                token=token,
                fee=fee,
                nonce=self.nonce,
                valid_from=obj["valid_from"],
                valid_until=obj["valid_until"],
                eth_auth_data=obj["eth_auth_type"]
            )
            eth_signature = self.wallet.eth_signer.sign(change_pub_key.get_eth_tx_bytes())
            eth_auth_data = change_pub_key.get_auth_data(eth_signature.signature)
            change_pub_key.eth_auth_data = eth_auth_data

            zk_signature = self.wallet.zk_signer.sign_tx(change_pub_key)
            change_pub_key.signature = zk_signature
        else:
            change_pub_key = ChangePubKey(
                account_id=obj["accountId"],
                account=obj["account"],
                new_pk_hash=obj["newPkHash"],
                token=obj["fee_token"],
                fee=obj["fee"],
                nonce=self.nonce,
                eth_auth_data=obj["ethAuthData"],
                signature=obj["signature"],
                valid_from=obj["validFrom"],
                valid_until=obj["validUntil"]
            )
        self.nonce += 1
        return change_pub_key

    async def _process_withdraw(self, obj: dict):
        if not obj[self.IS_ENCODED_TRANSACTION]:
            account_id = await self.wallet.get_account_id()
            token = await self.wallet.resolve_token(obj["token"])

            fee = obj["fee"]
            if fee is None:
                fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.withdraw,
                                                                        obj["eth_address"],
                                                                        token.id)
                fee = fee.total_fee
            else:
                fee = token.from_decimal(fee)
            amount = token.from_decimal(obj["amount"])

            withdraw = Withdraw(account_id=account_id,
                                from_address=self.wallet.address(),
                                to_address=obj["eth_address"],
                                amount=amount,
                                fee=fee,
                                nonce=self.nonce,
                                valid_from=obj["valid_from"],
                                valid_until=obj["valid_until"],
                                token=token)
            zk_signature = self.wallet.zk_signer.sign_tx(withdraw)
            withdraw.signature = zk_signature
        else:
            token = await self.wallet.resolve_token(obj["token"])
            withdraw = Withdraw(account_id=obj["accountId"],
                                from_address=obj["from"],
                                to_address=obj["to"],
                                amount=obj["amount"],
                                fee=obj["fee"],
                                nonce=self.nonce,
                                valid_from=obj["validFrom"],
                                valid_until=obj["validUntil"],
                                token=token,
                                signature=obj["signature"]
                                )
        self.nonce += 1
        return withdraw

    async def _process_transfer(self, obj):
        if not obj[self.IS_ENCODED_TRANSACTION]:
            account_id = await self.wallet.get_account_id()
            token = await self.wallet.resolve_token(obj["token"])

            fee = obj["fee"]
            if fee is None:
                fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.transfer,
                                                                        obj["to_address"],
                                                                        token.id)
                fee = fee.total_fee
            else:
                fee = token.from_decimal(fee)

            amount = token.from_decimal(obj["amount"])
            transfer = Transfer(
                account_id=account_id,
                from_address=obj["from_address"].lower(),
                to_address=obj["to_address"].lower(),
                token=token,
                amount=amount,
                fee=fee,
                nonce=self.nonce,
                valid_from=obj["valid_from"],
                valid_until=obj["valid_until"]
            )
            zk_signature = self.wallet.zk_signer.sign_tx(transfer)
            transfer.signature = zk_signature
        else:
            token = await self.wallet.resolve_token(obj["token"])
            transfer = Transfer(
                account_id=obj["accountId"],
                from_address=obj["from"],
                to_address=obj["to"],
                token=token,
                amount=obj["amount"],
                fee=obj["fee"],
                nonce=self.nonce,
                valid_from=obj["validFrom"],
                valid_until=obj["validUntil"],
                signature=obj["signature"]
            )
        self.nonce += 1
        return transfer

    async def _process_forced_exit(self, obj):
        if not obj[self.IS_ENCODED_TRANSACTION]:
            account_id = await self.wallet.get_account_id()
            token = await self.wallet.resolve_token(obj["token"])
            fee = obj["fee"]
            if fee is None:
                fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.transfer,
                                                                        obj["to_address"],
                                                                        token.id)
                fee = fee.total_fee
            else:
                fee = token.from_decimal(fee)
            forced_exit = ForcedExit(initiator_account_id=account_id,
                                     target=obj["target"],
                                     fee=fee,
                                     nonce=self.nonce,
                                     valid_from=obj["valid_from"],
                                     valid_until=obj["valid_until"],
                                     token=token)
            zk_signature = self.wallet.zk_signer.sign_tx(forced_exit)
            forced_exit.signature = zk_signature
        else:
            token = await self.wallet.resolve_token(obj["token"])
            forced_exit = ForcedExit(initiator_account_id=obj["initiatorAccountId"],
                                     target=obj["target"],
                                     fee=obj["fee"],
                                     nonce=self.nonce,
                                     valid_from=obj["valid_from"],
                                     valid_until=obj["valid_until"],
                                     token=token,
                                     signature=obj["signature"])
        self.nonce += 1
        return forced_exit

    async def _process_swap(self, obj):
        if not obj[self.IS_ENCODED_TRANSACTION]:
            fee_token = await self.wallet.resolve_token(obj["fee_token"])
            fee = obj["fee"]
            if fee is None:
                fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.swap,
                                                                        self.wallet.address(),
                                                                        fee_token.id)
                fee = fee.total_fee
            else:
                fee = fee_token.from_decimal(fee)

            amounts = obj["amounts"]
            orders = obj["orders"]
            if amounts is None:
                amounts = (orders[0].amount, orders[1].amount)
            else:
                amounts = (
                    orders[0].token_sell.from_decimal(amounts[0]),
                    orders[1].token_sell.from_decimal(amounts[1])
                )
            account_id = await self.wallet.get_account_id()
            swap = Swap(
                orders=orders,
                fee_token=fee_token,
                amounts=amounts,
                fee=fee,
                nonce=self.nonce,
                submitter_id=account_id,
                submitter_address=self.wallet.address()
            )
            swap.signature = self.wallet.zk_signer.sign_tx(swap)
        else:
            fee_token = await self.wallet.resolve_token(obj["feeToken"])
            swap = Swap(
                orders=obj["orders"],
                fee_token=fee_token,
                amounts=obj["amounts"],
                fee=obj["fee"],
                nonce=self.nonce,
                submitter_id=obj["submitterId"],
                submitter_address=obj["submitterAddress"],
                signature=obj["signature"]
            )
        self.nonce += 1
        return swap

    async def _process_mint_nft(self, obj):
        if not obj[self.IS_ENCODED_TRANSACTION]:
            fee_token = await self.wallet.resolve_token(obj["fee_token"])
            account_id = await self.wallet.get_account_id()

            fee = obj["fee"]
            if fee is None:
                fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.mint_nft,
                                                                        obj["recipient"],
                                                                        fee_token.id)
                fee = fee.total_fee
            else:
                fee = fee_token.from_decimal(fee)
            mint_nft = MintNFT(creator_id=account_id,
                               creator_address=self.wallet.address(),
                               content_hash=obj["content_hash"],
                               recipient=obj["recipient"],
                               fee=fee,
                               fee_token=fee_token,
                               nonce=self.nonce)
            zk_signature = self.wallet.zk_signer.sign_tx(mint_nft)
            mint_nft.signature = zk_signature
        else:
            fee_token = await self.wallet.resolve_token(obj["fee_token"])
            mint_nft = MintNFT(creator_id=obj["creatorId"],
                               creator_address=obj["creatorAddress"],
                               content_hash=obj["content_hash"],
                               recipient=obj["recipient"],
                               fee=obj["fee"],
                               fee_token=fee_token,
                               nonce=self.nonce,
                               signature=obj["signature"]
                               )
        self.nonce += 1
        return mint_nft

    async def _process_withdraw_nft(self, obj):
        if not obj[self.IS_ENCODED_TRANSACTION]:
            fee_token = await self.wallet.resolve_token(obj["fee_token"])

            fee = obj["fee"]
            if fee is None:
                fee = await self.wallet.zk_provider.get_transaction_fee(FeeTxType.withdraw_nft,
                                                                        obj["to_address"],
                                                                        fee_token.id
                                                                        )
                fee = fee.total_fee
            else:
                fee = fee_token.from_decimal(fee)

            account_id = await self.wallet.get_account_id()

            withdraw_nft = WithdrawNFT(
                account_id=account_id,
                from_address=self.wallet.address(),
                to_address=obj["to_address"],
                fee_token=fee_token,
                fee=fee,
                nonce=self.nonce,
                valid_from=obj["valid_from"],
                valid_until=obj["valid_until"],
                token_id=obj["nft_token"].id
            )
            zk_signature = self.wallet.zk_signer.sign_tx(withdraw_nft)
            withdraw_nft.signature = zk_signature
        else:
            fee_token = await self.wallet.resolve_token(obj["feeToken"])
            withdraw_nft = WithdrawNFT(
                account_id=obj["accountId"],
                from_address=obj["from"],
                to_address=obj["to"],
                fee_token=fee_token,
                fee=obj["fee"],
                nonce=obj["nonce"],
                valid_from=obj["validFrom"],
                valid_until=obj["validUntil"],
                token_id=obj["nft_token"].id,
                signature=obj["signature"]
            )
        self.nonce += 1
        return withdraw_nft

    async def _process_transactions(self):
        message = ""
        trs = []
        total_fee_map = dict()
        for obj in self.transactions:
            if obj[self.ENCODED_TRANSACTION_TYPE] == EncodedTxType.CHANGE_PUB_KEY:
                tr = await self._process_change_pub_key(obj)

                prev_value = total_fee_map.get(tr.token.symbol, Decimal(0))
                dec_fee = tr.token.decimal_amount(tr.fee)
                total_fee_map[tr.token.symbol] = dec_fee + prev_value

                message += tr.batch_message_part()
                trs.append(TransactionWithOptionalSignature(tr))
            elif obj[self.ENCODED_TRANSACTION_TYPE] == EncodedTxType.TRANSFER:
                tr = await self._process_transfer(obj)

                prev_value = total_fee_map.get(tr.token.symbol, Decimal(0))
                dec_fee = tr.token.decimal_amount(tr.fee)
                total_fee_map[tr.token.symbol] = dec_fee + prev_value

                message += tr.batch_message_part()
                trs.append(TransactionWithOptionalSignature(tr))
            elif obj[self.ENCODED_TRANSACTION_TYPE] == EncodedTxType.WITHDRAW:
                tr = await self._process_withdraw(obj)

                prev_value = total_fee_map.get(tr.token.symbol, Decimal(0))
                dec_fee = tr.token.decimal_amount(tr.fee)
                total_fee_map[tr.token.symbol] = dec_fee + prev_value

                message += tr.batch_message_part()
                trs.append(TransactionWithOptionalSignature(tr))
            elif obj[self.ENCODED_TRANSACTION_TYPE] == EncodedTxType.FORCED_EXIT:
                tr = await self._process_forced_exit(obj)

                prev_value = total_fee_map.get(tr.token.symbol, Decimal(0))
                dec_fee = tr.token.decimal_amount(tr.fee)
                total_fee_map[tr.token.symbol] = dec_fee + prev_value

                message += tr.batch_message_part()
                trs.append(TransactionWithOptionalSignature(tr))
            elif obj[self.ENCODED_TRANSACTION_TYPE] == EncodedTxType.MINT_NFT:
                tr = await self._process_mint_nft(obj)

                prev_value = total_fee_map.get(tr.fee_token.symbol, Decimal(0))
                dec_fee = tr.fee_token.decimal_amount(tr.fee)
                total_fee_map[tr.fee_token.symbol] = dec_fee + prev_value

                message += tr.batch_message_part()
                trs.append(TransactionWithOptionalSignature(tr))
            elif obj[self.ENCODED_TRANSACTION_TYPE] == EncodedTxType.WITHDRAW_NFT:
                tr = await self._process_withdraw_nft(obj)

                prev_value = total_fee_map.get(tr.fee_token.symbol, Decimal(0))
                dec_fee = tr.fee_token.decimal_amount(tr.fee)
                total_fee_map[tr.fee_token.symbol] = dec_fee + prev_value

                message += tr.batch_message_part()
                trs.append(TransactionWithOptionalSignature(tr))
            elif obj[self.ENCODED_TRANSACTION_TYPE] == EncodedTxType.SWAP:
                tr = await self._process_swap(obj)

                prev_value = total_fee_map.get(tr.fee_token.symbol, Decimal(0))
                dec_fee = tr.fee_token.decimal_amount(tr.fee)
                total_fee_map[tr.fee_token.symbol] = dec_fee + prev_value
                message += tr.batch_message_part()
                trs.append(TransactionWithOptionalSignature(tr, [None,
                                                                 tr.orders[0].eth_signature,
                                                                 tr.orders[1].eth_signature]
                                                            ))
            else:
                raise TypeError("_process_transactions is trying to process unimplemented type")
        message += f"Nonce: {self.batch_nonce}"
        result = dict(trans=trs, msg=message, total_fee=total_fee_map)
        return result



class ZkSyncProviderError(Exception):
    pass


class AccountDoesNotExist(ZkSyncProviderError):
    def __init__(self, address, *args):
        self.address = address
        super().__init__(*args)



from decimal import Decimal
from fractions import Fraction
from unittest import IsolatedAsyncioTestCase
from zksync_sdk.zksync_provider.types import FeeTxType
from zksync_sdk.types.responses import Fee
import asyncio
from web3 import Account, HTTPProvider, Web3

from zksync_sdk import (EthereumProvider, EthereumSignerWeb3, HttpJsonRPCTransport, Wallet, ZkSync,
                        ZkSyncLibrary, ZkSyncProviderV01, ZkSyncSigner, )
from zksync_sdk.zksync_provider.batch_builder import BatchBuilder
from zksync_sdk.network import rinkeby
from zksync_sdk.types import ChangePubKeyEcdsa, Token, TransactionWithSignature, \
    TransactionWithOptionalSignature, RatioType, Transfer, AccountTypes
from zksync_sdk.zksync_provider.transaction import TransactionStatus
from zksync_sdk.wallet import DEFAULT_VALID_FROM, DEFAULT_VALID_UNTIL


class TestWallet(IsolatedAsyncioTestCase):
    # 0x995a8b7f96cb837533b79775b6209696d51f435c
    private_key = "0xa045b52470d306ff78e91b0d2d92f90f7504189125a46b69423dc673fd6b4f3e"
    private_keys = [
        # 0x800455ca06265d0cf742086663a527d7c08049fc
        "0x601b47729b2820e94bc10125edc8d534858827428b449175a275069dc00c303f",
        # 0x3aa03b5bcba43eebcb98432507474ffb3423ac94
        "0xa7adf8459b4c9a62f09e0e5390983c0145fa20e88c9e5bf837d8bf3dcd05bd9c",
    ]
    receiver_address = "0x21dDF51966f2A66D03998B0956fe59da1b3a179F"
    forced_exit_account_address = "0x21dDF51966f2A66D03998B0956fe59da1b3aFFFE"
    nft_transfer_account_address = "0x995a8b7f96cb837533b79775b6209696d51f435c"

    async def get_wallet(self, private_key: str) -> Wallet:
        account = Account.from_key(private_key)
        ethereum_signer = EthereumSignerWeb3(account=account)

        w3 = Web3(HTTPProvider(
            endpoint_uri="https://rinkeby.infura.io/v3/bcf42e619a704151a1b0d95a35cb2e62"))
        provider = ZkSyncProviderV01(provider=HttpJsonRPCTransport(network=rinkeby))
        address = await provider.get_contract_address()
        zksync = ZkSync(account=account, web3=w3, zksync_contract_address=address.main_contract)
        ethereum_provider = EthereumProvider(w3, zksync)
        signer = ZkSyncSigner.from_account(account, self.library, rinkeby.chain_id)

        return Wallet(ethereum_provider=ethereum_provider, zk_signer=signer,
                      eth_signer=ethereum_signer, provider=provider)

    async def asyncSetUp(self):
        self.library = ZkSyncLibrary()
        self.wallet = await self.get_wallet(self.private_key)
        self.wallets = [await self.get_wallet(key) for key in self.private_keys]

    async def test_get_account_state(self):
        data = await self.wallet.zk_provider.get_state(self.wallet.address())
        assert data.address.lower() == self.wallet.address().lower()

    async def test_deposit(self):
        token = await self.wallet.resolve_token("USDT")
        await self.wallet.ethereum_provider.approve_deposit(token, Decimal(1))

        res = await self.wallet.ethereum_provider.deposit(token, Decimal(1),
                                                          self.wallet.address())
        assert res

    async def test_change_pubkey(self):
        trans = await self.wallet.set_signing_key("ETH", eth_auth_data=ChangePubKeyEcdsa())
        try:
            result = await trans.await_committed(attempts=1000, attempts_timeout=1000)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, str(ex)

    async def test_is_public_key_onset(self):
        pubkey_hash = self.wallet.zk_signer.pubkey_hash()
        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())
        await self.wallet.ethereum_provider.set_auth_pubkey_hash(pubkey_hash, nonce)
        assert await self.wallet.ethereum_provider.is_onchain_auth_pubkey_hash_set(nonce)

    async def test_transfer(self):
        tr = await self.wallet.transfer(self.receiver_address,
                                        amount=Decimal("0.01"), token="USDC")
        try:
            result = await tr.await_committed(attempts=20, attempts_timeout=100)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, str(ex)

    async def test_swap(self):
        order1 = await self.wallet.get_order('USDT', 'ETH', Fraction(1500, 1), RatioType.token, Decimal('1.0'))
        order2 = await self.wallets[0].get_order('ETH', 'USDT', Fraction(1, 1200), RatioType.token, Decimal('0.0007'))
        tr = await self.wallet.swap((order1, order2), 'ETH')
        try:
            result = await tr.await_committed(attempts=100, attempts_timeout=100)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_swap, getting status raises error: {ex}"

    async def test_batch(self):
        trs = []
        eth_token = await self.wallet.resolve_token("ETH")
        fee = (await self.wallet.zk_provider.get_transaction_fee(
            FeeTxType.transfer, self.receiver_address, "ETH"
        )).total_fee
        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())

        for i in range(3):
            tr, sig = await self.wallet.build_transfer(
                self.receiver_address,
                amount=1, token=eth_token, fee=fee, nonce=nonce + i)
            trs.append(TransactionWithSignature(tr, sig))
        res = await self.wallet.send_txs_batch(trs)
        self.assertEqual(len(res), 3)
        for i, tr in enumerate(res):
            try:
                result = await tr.await_committed(attempts=100, attempts_timeout=500)
                self.assertEqual(result.status, TransactionStatus.COMMITTED)
            except Exception as ex:
                assert False, f"test_batch, getting transaction {i}  result has failed with error: {ex}"

    async def test_build_batch_transfer(self):
        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())
        builder = BatchBuilder.from_wallet(self.wallet, nonce)
        for i in range(2):
            builder.add_transfer(self.receiver_address, "ETH", Decimal("0.00005"))
        build_result = await builder.build()
        print(f"Total fees: {build_result.total_fees}")
        transactions = await self.wallet.zk_provider.submit_batch_builder_txs_batch(build_result.transactions,
                                                                                    build_result.signature)
        for i, tran in enumerate(transactions):
            try:
                result = await tran.await_committed(attempts=1000, attempts_timeout=1000)
                self.assertEqual(result.status, TransactionStatus.COMMITTED)
            except Exception as ex:
                assert False, f"test_build_batch_transfer, transaction {i} " \
                              f"has failed with error: {ex}"

    async def test_build_batch_change_pub_key(self):
        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())
        builder = BatchBuilder.from_wallet(self.wallet, nonce)
        builder.add_change_pub_key("ETH", eth_auth_type=ChangePubKeyEcdsa())
        builder.add_transfer(self.receiver_address, "USDT", Decimal("0.001"))
        build_result = await builder.build()
        print(f"Total fees: {build_result.total_fees}")
        transactions = await self.wallet.zk_provider.submit_batch_builder_txs_batch(build_result.transactions,
                                                                                    build_result.signature)
        self.assertEqual(len(transactions), 2)
        for i, tran in enumerate(transactions):
            try:
                result = await tran.await_committed(attempts=100, attempts_timeout=1000)
                self.assertEqual(result.status, TransactionStatus.COMMITTED)
            except Exception as ex:
                assert False, f"test_build_batch_change_pub_key, transaction {i} " \
                              f"has failed with error: {ex}"

    async def test_build_batch_withdraw(self):
        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())
        builder = BatchBuilder.from_wallet(self.wallet, nonce)
        builder.add_withdraw(self.receiver_address,
                             "USDT",
                             Decimal("0.000001")
                             )
        build_result = await builder.build()
        print(f"Total fees: {build_result.total_fees}")
        transactions = await self.wallet.zk_provider.submit_batch_builder_txs_batch(build_result.transactions,
                                                                                    build_result.signature)
        self.assertEqual(len(transactions), 1)

        try:
            result = await transactions[0].await_committed(attempts=100, attempts_timeout=1000)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_build_batch_withdraw, transaction has failed with error: {ex}"

    async def test_build_batch_mint_nft(self):
        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())
        builder = BatchBuilder.from_wallet(self.wallet, nonce)
        builder.add_mint_nft("0x0000000000000000000000000000000000000000000000000000000000000123",
                             self.receiver_address,
                             "USDC"
                             )
        build_result = await builder.build()
        print(f"Total fees: {build_result.total_fees}")
        transactions = await self.wallet.zk_provider.submit_batch_builder_txs_batch(build_result.transactions,
                                                                                    build_result.signature)
        self.assertEqual(len(transactions), 1)

        try:
            result = await transactions[0].await_committed(attempts=1000, attempts_timeout=1000)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_build_batch_mint_nft, transaction has failed with error: {ex}"

    async def test_build_batch_withdraw_nft(self):
        account_state = await self.wallet.get_account_state()
        nfts = account_state.verified.nfts.values()
        if not nfts:
            return
        nfts_iterator = iter(nfts)
        first_value = next(nfts_iterator)

        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())
        builder = BatchBuilder.from_wallet(self.wallet, nonce)
        builder.add_withdraw_nft(self.receiver_address,
                                 first_value,
                                 "USDC"
                                 )
        build_result = await builder.build()
        print(f"Total fees: {build_result.total_fees}")
        transactions = await self.wallet.zk_provider.submit_batch_builder_txs_batch(build_result.transactions,
                                                                                    build_result.signature)
        self.assertEqual(len(transactions), 1)
        try:
            result = await transactions[0].await_committed(attempts=1000, attempts_timeout=1000)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_build_batch_withdraw_nft, transaction has failed with error: {ex}"

    async def test_build_batch_swap(self):
        nonce = await self.wallet.zk_provider.get_account_nonce(self.wallet.address())
        nonce0 = await self.wallets[0].zk_provider.get_account_nonce(self.wallets[0].address())
        builder = BatchBuilder.from_wallet(self.wallet, nonce)
        test_n = 2
        for i in range(test_n):
            order1 = await self.wallet.get_order('USDT',
                                                 'ETH',
                                                 Fraction(1500, 1),
                                                 RatioType.token,
                                                 Decimal('0.1')
                                                 , nonce=nonce + i
                                                 )
            order2 = await self.wallets[0].get_order('ETH',
                                                     'USDT',
                                                     Fraction(1, 1200),
                                                     RatioType.token,
                                                     Decimal('0.00007'),
                                                     nonce=nonce0 + i)
            builder.add_swap((order1, order2), 'ETH')
        build_result = await builder.build()
        print(f"Total fees: {build_result.total_fees}")
        transactions = await self.wallet.zk_provider.submit_batch_builder_txs_batch(build_result.transactions,
                                                                                    build_result.signature)
        self.assertEqual(len(transactions), test_n)
        for i, tran in enumerate(transactions):
            try:
                result = await tran.await_committed(attempts=1000, attempts_timeout=1000)
                self.assertEqual(result.status, TransactionStatus.COMMITTED)
            except Exception as ex:
                assert False, f"test_build_batch_swap, transaction {i} " \
                              f"has failed with error: {ex}"

    async def test_forced_exit(self):
        result_transaction = await self.wallet.transfer(self.forced_exit_account_address, Decimal("0.1"), "USDC")
        result = await result_transaction.await_committed()
        self.assertEqual(result.status, TransactionStatus.COMMITTED)
        tr = await self.wallet.forced_exit(self.forced_exit_account_address, "USDC")
        try:
            result = await tr.await_verified(attempts=10, attempts_timeout=1000)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_forced_exit, getting transaction result has failed with error: {result.error_message}"

    async def test_mint_nft(self):
        tr = await self.wallet.mint_nft("0x0000000000000000000000000000000000000000000000000000000000000123",
                                        self.receiver_address, "USDC")
        try:
            result = await tr.await_committed(attempts=20, attempts_timeout=100)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_mint_nft, getting transaction result has failed with error: {ex}"

    async def test_transfer_nft(self):
        """
        INFO: During the testing there are cases when this wallet does not own any NFT tokens by default,
              use mint_nft to VERIFIED state took too long and failed
              There are 2 solutions for the whole situation:
              1. Prepare the docker with local ZkSync & Eth servers & achieve VERIFIED state fast =>
                 Any token or data can be transfered/deposited inside the test and do manipulations
              2. If this wallet does not have NFT tokens do nothing
                 Currently this choise is made

              PS: previous version of the tests was passing due to no one does not test the trasaction result
                  it failed
        """

        account_state = await self.wallet.zk_provider.get_state(self.nft_transfer_account_address)
        nfts = account_state.verified.nfts.items()
        first_value = None
        for key, value in nfts:
            if value.content_hash == "0x0000000000000000000000000000000000000000000000000000000000000123":
                first_value = value
                break
        if first_value is None:
            return

        txs = await self.wallet.transfer_nft(
            self.wallet.address(),
            first_value,
            "USDC",
            Decimal(0.01)
        )
        self.assertEqual(len(txs), 2)
        for i, tr in enumerate(txs):
            try:
                result = await tr.await_committed(attempts=1000, attempts_timeout=1000)
                self.assertEqual(result.status, TransactionStatus.COMMITTED)
            except Exception as ex:
                assert False, f"test_transfer_nft, transaction {i} has failed with error: {ex}"

    async def test_withdraw_nft(self):
        """
        INFO: During the testing there are cases when this wallet does not own any NFT tokens by default,
              use mint_nft to VERIFIED state took too long and failed
              There are 2 solutions for the whole situation:
              1. Prepare the docker with local ZkSync & Eth servers & achieve VERIFIED state fast =>
                 Any token or data can be transfered/deposited inside the test and do manipulations
              2. If this wallet does not have NFT tokens do nothing
                 Currently this choise is made

              PS: previous version of the tests was passing due to no one does not test the trasaction result
                  it failed
        """
        account_state = await self.wallet.zk_provider.get_state(self.wallet.address())

        nfts = account_state.verified.nfts.values()
        if not nfts:
            return
        nfts_iter = iter(nfts)
        first_value = next(nfts_iter)
        tr = await self.wallet.withdraw_nft(self.nft_transfer_account_address, first_value, "USDC")
        try:
            result = await tr.await_committed(attempts=1000, attempts_timeout=1000)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_withdraw_nft, transaction has failed with error: {ex}"

    async def test_withdraw(self):
        tr = await self.wallet.withdraw(self.receiver_address,
                                        Decimal("0.000001"), "USDT")
        try:
            result = await tr.await_committed(attempts=30, attempts_timeout=100)
            self.assertEqual(result.status, TransactionStatus.COMMITTED)
        except Exception as ex:
            assert False, f"test_withdraw, transaction has failed with error: {ex}"

    async def test_get_tokens(self):
        tokens = await self.wallet.zk_provider.get_tokens()
        assert tokens.find_by_symbol("ETH")

    async def test_is_signing_key_set(self):
        assert await self.wallet.is_signing_key_set()

    async def test_toggle_2fa(self):
        """
        Relate to the server-side code it must be Owned type if enable_2fa is passed
        let new_type = if toggle_2fa.enable {
            EthAccountType::Owned
        } else {
            EthAccountType::No2FA
        };
        """
        result = await self.wallet.enable_2fa()
        self.assertTrue(result)
        account_state = await self.wallet.get_account_state()
        self.assertEqual(AccountTypes.OWNED, account_state.account_type)

        pub_key_hash = self.wallet.zk_signer.pubkey_hash_str()
        result = await self.wallet.disable_2fa(pub_key_hash)
        self.assertTrue(result)
        account_state = await self.wallet.get_account_state()
        self.assertEqual(AccountTypes.NO_2FA, account_state.account_type)


class TestEthereumProvider(IsolatedAsyncioTestCase):
    private_key = "0xa045b52470d306ff78e91b0d2d92f90f7504189125a46b69423dc673fd6b4f3e"

    async def asyncSetUp(self) -> None:
        self.account = Account.from_key(self.private_key)
        self.library = ZkSyncLibrary()

        w3 = Web3(HTTPProvider(
            endpoint_uri="https://rinkeby.infura.io/v3/bcf42e619a704151a1b0d95a35cb2e62"))
        provider = ZkSyncProviderV01(provider=HttpJsonRPCTransport(network=rinkeby))
        address = await provider.get_contract_address()
        self.zksync = ZkSync(account=self.account, web3=w3,
                             zksync_contract_address=address.main_contract)
        self.ethereum_provider = EthereumProvider(w3, self.zksync)

    async def test_approve_deposit(self):
        token = Token(
            address=Web3.toChecksumAddress('0xeb8f08a975ab53e34d8a0330e0d34de942c95926'),
            id=20, symbol='USDC',
            decimals=18)
        assert await self.ethereum_provider.approve_deposit(token, Decimal(1))

    async def test_full_exit(self):
        token = Token(
            address=Web3.toChecksumAddress('0xD2084eA2AE4bBE1424E4fe3CDE25B713632fb988'),
            id=20, symbol='BAT',
            decimals=18)
        assert await self.ethereum_provider.full_exit(token, 6713)

    async def test_full_exit_nft(self):
        """
        INFO: made by getting all NFT by corresponded address & dumping,
                Symbol: 'NFT-70848'
                '70848'
                address: '0x5e71f0f9b891f22d79ff8697dd4e3e0db371cda5'
                creator_address: '0x995a8b7f96cb837533b79775b6209696d51f435c'
                id: 70848
                account_id: 36357
        """
        account_id = 36357
        token = Token(
            address=Web3.toChecksumAddress('0x5e71f0f9b891f22d79ff8697dd4e3e0db371cda5'),
            id=70848,
            symbol='NFT-70848',
            decimals=0
        )
        assert await self.ethereum_provider.full_exit_nft(token, account_id)

    async def test_is_deposit_approved(self):
        token = Token(
            address=Web3.toChecksumAddress('0xD2084eA2AE4bBE1424E4fe3CDE25B713632fb988'),
            id=20, symbol='BAT',
            decimals=18)
        assert await self.ethereum_provider.is_deposit_approved(token, 1)


class TestZkSyncProvider(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.provider = ZkSyncProviderV01(provider=HttpJsonRPCTransport(network=rinkeby))

    async def test_get_token_price(self):
        tokens = await self.provider.get_tokens()
        price = await self.provider.get_token_price(tokens.find_by_symbol("USDC"))
        self.assertAlmostEqual(float(price), 1.0, delta=0.2)



from dataclasses import asdict
from decimal import Decimal
from typing import List, Optional, Union
from web3 import Web3

from zksync_sdk.types import (AccountState, ContractAddress, EncodedTx, EthOpInfo, Fee, Token,
                              TokenLike, Tokens, TransactionDetails, TransactionWithSignature,
                              TransactionWithOptionalSignature,
                              TxEthSignature, Toggle2FA, )
from zksync_sdk.zksync_provider.error import AccountDoesNotExist
from zksync_sdk.zksync_provider.interface import ZkSyncProviderInterface
from zksync_sdk.zksync_provider.types import FeeTxType
from zksync_sdk.zksync_provider.transaction import Transaction

__all__ = ['ZkSyncProviderV01']


class ZkSyncProviderV01(ZkSyncProviderInterface):
    async def submit_tx(self, tx: EncodedTx, signature: Union[Optional[TxEthSignature], List[Optional[TxEthSignature]]],
                        fast_processing: bool = False) -> Transaction:
        if isinstance(signature, List):
            signature = [s.dict() if s is not None else None for s in signature]
        else:
            signature = signature.dict() if signature is not None else None
        trans_id = await self.provider.request("tx_submit",
                                               [tx.dict(), signature, fast_processing])
        return Transaction.build_transaction(self, trans_id)

    async def get_tokens(self) -> Tokens:
        data = await self.provider.request("tokens", None)
        tokens = [Token(address=Web3.toChecksumAddress(token['address']),
                        id=token['id'],
                        symbol=token['symbol'],
                        decimals=token['decimals']
                        ) for token in data.values()]
        return Tokens(tokens=tokens)

    async def submit_txs_batch(self, transactions: List[TransactionWithSignature],
                               signatures: Optional[
                                   Union[List[TxEthSignature], TxEthSignature]
                               ] = None) -> List[Transaction]:
        if signatures is None:
            signatures = []
        elif isinstance(signatures, TxEthSignature):
            signatures = [signatures]
        transactions = [tr.dict() for tr in transactions]
        signatures = [sig.dict() for sig in signatures]
        trans_ids: List[str] = await self.provider.request("submit_txs_batch", [transactions, signatures])
        return [Transaction.build_transaction(self, trans_id) for trans_id in trans_ids]

    async def submit_batch_builder_txs_batch(self, transactions: List[TransactionWithOptionalSignature],
                                             signature: TxEthSignature) -> List[Transaction]:
        trans = [tr.dict() for tr in transactions]
        params = [trans, signature.dict()]
        trans_ids: List[str] = await self.provider.request("submit_txs_batch", params)
        return [Transaction.build_transaction(self, trans_id) for trans_id in trans_ids]

    async def get_contract_address(self) -> ContractAddress:
        data = await self.provider.request("contract_address", None)
        return ContractAddress(**data)

    async def get_state(self, address: str) -> AccountState:
        data = await self.provider.request("account_info", [address])
        if data is None:
            raise AccountDoesNotExist(address=address)
        if "accountType" in data and isinstance(data["accountType"], dict) and \
                list(data["accountType"].keys())[0] == 'No2FA':
            data["accountType"] = 'No2FA'
        return AccountState(**data)

    async def get_confirmations_for_eth_op_amount(self) -> int:
        return await self.provider.request("get_confirmations_for_eth_op_amount", None)

    async def get_account_nonce(self, address: str) -> int:
        state = await self.get_state(address)
        return state.get_nonce()

    async def get_tx_receipt(self, address: str) -> TransactionDetails:
        return await self.provider.request("tx_info", [address])

    async def get_eth_tx_for_withdrawal(self, withdrawal_hash: str) -> str:
        return await self.provider.request("get_eth_tx_for_withdrawal", [withdrawal_hash])

    async def get_priority_op_status(self, serial_id: int) -> EthOpInfo:
        data = await self.provider.request("ethop_info", [serial_id])
        return EthOpInfo(**data)

    # Please note that the batch fee returns the fee of the transaction in int and not in Fee
    # This is a server-side feature
    async def get_transactions_batch_fee(self, tx_types: List[FeeTxType], addresses: List[str],
                                         token_like) -> int:

        data = await self.provider.request('get_txs_batch_fee_in_wei',
                                           [[tx_type.value for tx_type in tx_types],
                                            addresses, token_like])
        return int(data["totalFee"])

    async def get_transaction_fee(self, tx_type: FeeTxType, address: str,
                                  token_like: TokenLike) -> Fee:

        data = await self.provider.request('get_tx_fee', [tx_type.value, address, token_like])
        return Fee(**data)

    async def get_token_price(self, token: Token) -> Decimal:
        data = await self.provider.request('get_token_price', [token.symbol])
        return Decimal(data)

    async def toggle_2fa(self, toggle2fa: Toggle2FA) -> bool:
        data = await self.provider.request('toggle_2fa', [toggle2fa.dict()])
        return 'success' in data and data['success']



import asyncio
from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional


class TransactionStatus(Enum):
    FAILED = auto()
    COMMITTED = auto()
    VERIFIED = auto()


@dataclass
class TransactionResult:
    status: TransactionStatus
    fail_reason: str


class Transaction:

    @classmethod
    def build_transaction(cls, provider, transaction_id: str):
        transaction = cls(provider, transaction_id)
        return transaction

    def __init__(self, provider, transaction_hash: str):
        self.provider = provider
        self.transaction_hash = transaction_hash

    async def await_committed(self, attempts: Optional[int] = None, attempts_timeout: Optional[int] = None) \
            -> TransactionResult:
        status = TransactionResult(TransactionStatus.FAILED,
                                   f"Transaction has not been executed with amount of attempts {attempts}"
                                   f"and timeout {attempts_timeout}")
        while True:
            if attempts is not None:
                if attempts <= 0:
                    return status
            transaction_details = await self.provider.get_tx_receipt(self.transaction_hash)
            if attempts is not None:
                attempts -= 1
            if "failReason" in transaction_details and transaction_details["failReason"] is not None:
                return TransactionResult(TransactionStatus.FAILED, transaction_details['failReason'])

            if "block" in transaction_details:
                block = transaction_details["block"]
                if block is not None and "committed" in block and block["committed"]:
                    return TransactionResult(TransactionStatus.COMMITTED, "")
            if attempts_timeout is not None:
                await asyncio.sleep(attempts_timeout / 1000)

    async def await_verified(self, attempts: Optional[int] = None, attempts_timeout: Optional[int] = None):
        intermediate_status = TransactionResult(
            TransactionStatus.FAILED,
            f"Transaction has not been executed with amount of attempts {attempts}"
            f"and timeout {attempts_timeout}")
        while True:
            if attempts is not None:
                if attempts <= 0:
                    return intermediate_status

            transaction_details = await self.provider.get_tx_receipt(self.transaction_hash)
            if attempts is not None:
                attempts -= 1
            if "failReason" in transaction_details and transaction_details["failReason"] is not None:
                return TransactionResult(TransactionStatus.FAILED, transaction_details['failReason'])

            if "block" in transaction_details:
                block = transaction_details["block"]
                if block is not None and "committed" in block and block["committed"]:
                    intermediate_status = TransactionResult(TransactionStatus.COMMITTED, "")

            if "block" in transaction_details:
                block = transaction_details["block"]
                if block is not None and \
                        "verified" in block and \
                        block["verified"]:
                    return TransactionResult(TransactionStatus.VERIFIED, "")

            if attempts_timeout is not None:
                await asyncio.sleep(attempts_timeout / 1000)



from typing import List
from math import ceil

AMOUNT_EXPONENT_BIT_WIDTH = 5
AMOUNT_MANTISSA_BIT_WIDTH = 35
FEE_EXPONENT_BIT_WIDTH = 5
FEE_MANTISSA_BIT_WIDTH = 11
MAX_NUMBER_OF_ACCOUNTS = 2 ** 24
MAX_NUMBER_OF_TOKENS = 2 ** 32 - 1


class SerializationError(Exception):
    pass


class WrongIntegerError(SerializationError):
    pass


class WrongBitsError(SerializationError):
    pass


class ValueNotPackedError(SerializationError):
    pass


class WrongValueError(SerializationError):
    pass


def int_to_bytes(val: int, length=4):
    return val.to_bytes(length, byteorder='big')


def num_to_bits(integer: int, bits: int):
    """
    INFO: Can't be used without correct input data of value & corresponded amount of bits, take care
    """
    results = []
    for i in range(bits):
        results.append(integer & 1)
        integer //= 2
    return results


def integer_to_float(integer: int, exp_bits: int, mantissa_bits: int, exp_base: int) -> List[int]:
    max_exponent_power = 2 ** exp_bits - 1
    max_exponent = exp_base ** max_exponent_power
    max_mantissa = 2 ** mantissa_bits - 1
    if integer > max_mantissa * max_exponent:
        raise WrongIntegerError

    exponent = 0
    exponent_temp = 1
    while integer > max_mantissa * exponent_temp:
        exponent_temp = exponent_temp * exp_base
        exponent += 1
    mantissa = integer // exponent_temp
    if exponent != 0:
        variant1 = exponent_temp * mantissa
        variant2 = exponent_temp // exp_base * max_mantissa
        diff1 = integer - variant1
        diff2 = integer - variant2
        if diff2 < diff1:
            mantissa = max_mantissa
            exponent -= 1

    data = num_to_bits(exponent, exp_bits) + num_to_bits(mantissa, mantissa_bits)
    data = list(reversed(data))
    result = list(reversed(bits_into_bytes_in_be_order(data)))

    return result


def integer_to_float_up(integer: int, exp_bits: int, mantissa_bits: int, exp_base) -> List[int]:
    max_exponent_power = 2 ** exp_bits - 1
    max_exponent = exp_base ** max_exponent_power
    max_mantissa = 2 ** mantissa_bits - 1

    if integer > max_mantissa * max_exponent:
        raise WrongIntegerError("Integer is too big")

    exponent = 0
    exponent_temp = 1
    while integer > max_mantissa * exponent_temp:
        exponent_temp = exponent_temp * exp_base
        exponent += 1

    mantissa = int(ceil(integer / exponent_temp))
    encoding = num_to_bits(exponent, exp_bits) + num_to_bits(mantissa, mantissa_bits)
    data = list(reversed(encoding))
    result = list(reversed(bits_into_bytes_in_be_order(data)))

    return result


def bits_into_bytes_in_be_order(bits: List[int]):
    if len(bits) % 8 != 0:
        raise WrongBitsError("wrong number of bits")
    size = len(bits) // 8
    result = [0] * size
    for i in range(size):
        value = 0
        for j in range(8):
            value |= bits[i * 8 + j] * 2 ** (7 - j)
        result[i] = value

    return result


def reverse_bit(b):
    b = ((b & 0xf0) >> 4) | ((b & 0x0f) << 4)
    b = ((b & 0xcc) >> 2) | ((b & 0x33) << 2)
    b = ((b & 0xaa) >> 1) | ((b & 0x55) << 1)
    return b


def reverse_bits(buffer: List[int]):
    return list(reversed(buffer))


def buffer_to_bits_be(buff):
    res = [0] * len(buff) * 8
    for i, b in enumerate(buff):
        for j in range(8):
            res[i * 8 + j] = (b >> (7 - j)) & 1
    return res


def pack_fee(amount: int):
    return bytes(reverse_bits(
        integer_to_float(amount, FEE_EXPONENT_BIT_WIDTH, FEE_MANTISSA_BIT_WIDTH, 10)
    ))


def pack_amount(amount: int) -> bytes:
    return bytes(reverse_bits(
        integer_to_float(amount, AMOUNT_EXPONENT_BIT_WIDTH, AMOUNT_MANTISSA_BIT_WIDTH, 10)
    ))


def float_to_integer(float_bytes: bytes, exp_bits, mantissa_bits, exp_base_number):
    bits = list(reversed(buffer_to_bits_be(list(float_bytes))))
    exponent = 0
    exp_pow2 = 1
    for i in range(exp_bits):
        if bits[i] == 1:
            exponent += exp_pow2
        exp_pow2 *= 2
    exponent = exp_base_number ** exponent
    mantissa = 0
    mantissa_pow2 = 1
    for i in range(exp_bits, exp_bits + mantissa_bits):
        if bits[i] == 1:
            mantissa += mantissa_pow2
        mantissa_pow2 *= 2

    return exponent * mantissa


def pack_amount_up(amount: int):
    return bytes(reverse_bits(
        integer_to_float_up(amount, AMOUNT_EXPONENT_BIT_WIDTH, AMOUNT_MANTISSA_BIT_WIDTH, 10)
    ))


def pack_fee_up(fee: int):
    return bytes(reverse_bits(
        integer_to_float_up(fee, FEE_EXPONENT_BIT_WIDTH, FEE_MANTISSA_BIT_WIDTH, 10)
    ))


def closest_packable_amount(amount: int) -> int:
    packed_amount = pack_amount(amount)
    return float_to_integer(
        packed_amount,
        AMOUNT_EXPONENT_BIT_WIDTH,
        AMOUNT_MANTISSA_BIT_WIDTH,
        10
    )


def closest_greater_or_eq_packable_amount(amount: int) -> int:
    packed_amount = pack_amount_up(amount)
    return float_to_integer(packed_amount, AMOUNT_EXPONENT_BIT_WIDTH, AMOUNT_MANTISSA_BIT_WIDTH, 10)


def closest_packable_transaction_fee(fee: int) -> int:
    packed_fee = pack_fee(fee)
    return float_to_integer(
        packed_fee,
        FEE_EXPONENT_BIT_WIDTH,
        FEE_MANTISSA_BIT_WIDTH,
        10
    )


def closest_greater_or_eq_packable_fee(fee: int) -> int:
    packed_fee = pack_fee_up(fee)
    return float_to_integer(packed_fee, FEE_EXPONENT_BIT_WIDTH, FEE_MANTISSA_BIT_WIDTH, 10)


def packed_fee_checked(fee: int):
    if closest_packable_transaction_fee(fee) != fee:
        raise ValueNotPackedError
    return pack_fee(fee)


def packed_amount_checked(amount: int):
    if closest_packable_amount(amount) != amount:
        raise ValueNotPackedError
    return pack_amount(amount)


def serialize_nonce(nonce: int):
    if nonce < 0:
        raise WrongValueError
    return int_to_bytes(nonce, 4)


def serialize_timestamp(timestamp: int):
    if timestamp < 0:
        raise WrongValueError
    return b"\x00" * 4 + int_to_bytes(timestamp, 4)


def serialize_token_id(token_id: int):
    if token_id < 0:
        raise WrongValueError
    if token_id > MAX_NUMBER_OF_TOKENS:
        raise WrongValueError
    return int_to_bytes(token_id, 4)


def serialize_account_id(account_id: int):
    if account_id < 0:
        raise WrongValueError
    if account_id > MAX_NUMBER_OF_ACCOUNTS:
        raise WrongValueError
    return int_to_bytes(account_id, 4)


def remove_address_prefix(address: str) -> str:
    if address.startswith('0x'):
        return address[2:]

    if address.startswith('sync:'):
        return address[5:]

    return address


def serialize_address(address: str) -> bytes:
    address = remove_address_prefix(address)
    address_bytes = bytes.fromhex(address)
    if len(address_bytes) != 20:
        raise WrongValueError
    return address_bytes


def serialize_content_hash(content_hash: str) -> bytes:
    if content_hash.startswith('0x'):
        content_hash = content_hash[2:]
    return bytes.fromhex(content_hash)


def serialize_ratio_part(part: int) -> bytes:
    # turn the number into bytes and 0-pad to length 15
    return bytes.fromhex(hex(part)[2:].zfill(15 * 2))



from unittest import TestCase
from fractions import Fraction

from web3 import Account

from zksync_sdk import ZkSyncLibrary, EthereumSignerWeb3
from zksync_sdk.serializers import closest_packable_amount, closest_packable_transaction_fee
from zksync_sdk.types import ChainId, ForcedExit, Token, Transfer, Withdraw, MintNFT, WithdrawNFT, Order, Swap, Tokens, \
    EncodedTxValidator
from zksync_sdk.zksync_signer import ZkSyncSigner

PRIVATE_KEY = "0x000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f"

import json


class ZkSyncSignerTest(TestCase):
    def setUp(self):
        self.library = ZkSyncLibrary()

    def test_derive_pub_key(self):
        account = Account.from_key(PRIVATE_KEY)
        signer = ZkSyncSigner.from_account(account, self.library, ChainId.MAINNET)
        assert signer.public_key.hex() == "40771354dc314593e071eaf4d0f42ccb1fad6c7006c57464feeb7ab5872b7490"

    def test_transfer_bytes(self):
        tr = Transfer(from_address="0xedE35562d3555e61120a151B3c8e8e91d83a378a",
                      to_address="0x19aa2ed8712072e918632259780e587698ef58df",
                      token=Token.eth(),
                      amount=1000000000000, fee=1000000, nonce=12, valid_from=0,
                      valid_until=4294967295, account_id=44)
        res = "fa010000002cede35562d3555e61120a151b3c8e8e91d83a378a19aa2ed8712072e918632259780e587698ef58df000000004a817c80027d030000000c000000000000000000000000ffffffff"
        assert tr.encoded_message().hex() == res

    def test_withdraw_bytes(self):
        tr = Withdraw(from_address="0xedE35562d3555e61120a151B3c8e8e91d83a378a",
                      to_address="0x19aa2ed8712072e918632259780e587698ef58df",
                      token=Token.eth(),
                      amount=1000000000000, fee=1000000, nonce=12, valid_from=0,
                      valid_until=4294967295, account_id=44)

        res = "fc010000002cede35562d3555e61120a151b3c8e8e91d83a378a19aa2ed8712072e918632259780e587698ef58df000000000000000000000000000000e8d4a510007d030000000c000000000000000000000000ffffffff"
        assert tr.encoded_message().hex() == res

    def test_order_bytes(self):
        token1 = Token.eth()
        token2 = Token(id=2, symbol='', address='', decimals=0)  # only id matters
        order = Order(account_id=6, nonce=18, token_sell=token1, token_buy=token2,
                      ratio=Fraction(1, 2), amount=1000000,
                      recipient='0x823b6a996cea19e0c41e250b20e2e804ea72ccdf',
                      valid_from=0, valid_until=4294967295)
        res = '6f0100000006823b6a996cea19e0c41e250b20e2e804ea72ccdf0000001200000000000000020000000000000000000000000000010000000000000000000000000000020001e84800000000000000000000000000ffffffff'
        assert order.encoded_message().hex() == res

    def test_swap_bytes(self):
        token1 = Token(id=1, symbol='', address='', decimals=0)  # only id matters
        token2 = Token(id=2, symbol='', address='', decimals=0)  # only id matters
        token3 = Token(id=3, symbol='', address='', decimals=0)  # only id matters
        order1 = Order(account_id=6, nonce=18, token_sell=token1, token_buy=token2,
                       ratio=Fraction(1, 2), amount=1000000,
                       recipient='0x823b6a996cea19e0c41e250b20e2e804ea72ccdf',
                       valid_from=0, valid_until=4294967295)
        order2 = Order(account_id=44, nonce=101, token_sell=token2, token_buy=token1,
                       ratio=Fraction(3, 1), amount=2500000,
                       recipient='0x63adbb48d1bc2cf54562910ce54b7ca06b87f319',
                       valid_from=0, valid_until=4294967295)
        swap = Swap(orders=(order1, order2), nonce=1, amounts=(1000000, 2500000),
                    submitter_id=5, submitter_address="0xedE35562d3555e61120a151B3c8e8e91d83a378a",
                    fee_token=token3, fee=123)
        res = "f40100000005ede35562d3555e61120a151b3c8e8e91d83a378a000000017b1e76f6f124bae1917435a02cfbf5571d79ddb8380bc4bf4858c9e9969487000000030f600001e848000004c4b400"
        assert swap.encoded_message().hex() == res

    def test_order_deserialization(self):
        token1 = Token(id=1, symbol='', address='', decimals=0)  # only id matters
        token2 = Token(id=2, symbol='', address='', decimals=0)  # only id matters
        tokens = Tokens(tokens=[token1, token2])

        order = Order(account_id=7, nonce=18, token_sell=token1, token_buy=token2,
                      ratio=Fraction(1, 4), amount=1000000,
                      recipient='0x823b6a996cea19e0c41e250b20e2e804ea72ccdf',
                      valid_from=0, valid_until=4294967295)
        serialized_order = order.dict()
        from_json_order = Order.from_json(serialized_order, tokens)
        self.assertEqual(order.account_id, from_json_order.account_id)
        self.assertEqual(order.nonce, from_json_order.nonce)
        self.assertEqual(order.token_sell, from_json_order.token_sell)
        self.assertEqual(order.token_buy, from_json_order.token_buy)
        self.assertEqual(order.ratio, from_json_order.ratio)
        self.assertEqual(order.recipient, from_json_order.recipient)
        self.assertEqual(order.valid_from, from_json_order.valid_from)
        self.assertEqual(order.valid_until, from_json_order.valid_until)

    def test_order_zksync_signature_checking(self):
        account = Account.from_key(PRIVATE_KEY)
        signer = ZkSyncSigner.from_account(account, self.library, ChainId.MAINNET)

        token1 = Token(id=1, symbol='', address='', decimals=0)  # only id matters
        token2 = Token(id=2, symbol='', address='', decimals=0)  # only id matters
        tokens_pool = Tokens(tokens=[token1, token2])

        order = Order(account_id=7, nonce=18, token_sell=token1, token_buy=token2,
                      ratio=Fraction(1, 4), amount=1000000,
                      recipient='0x823b6a996cea19e0c41e250b20e2e804ea72ccdf',
                      valid_from=0, valid_until=4294967295)

        order.signature = signer.sign_tx(order)

        validator = EncodedTxValidator(self.library)
        serialized_order = json.dumps(order.dict(), indent=4)
        print(f"json : {serialized_order}")
        deserialized_order = Order.from_json(json.loads(serialized_order), tokens_pool)
        ret = validator.is_valid_signature(deserialized_order)
        self.assertTrue(ret)

    def test_is_valid_order_deserialized(self):
        account = Account.from_key(PRIVATE_KEY)
        zksync_signer = ZkSyncSigner.from_account(account, self.library, ChainId.MAINNET)
        ethereum_signer = EthereumSignerWeb3(account=account)

        token1 = Token(id=1, symbol='', address='', decimals=0)  # only id matters
        token2 = Token(id=2, symbol='', address='', decimals=0)  # only id matters
        tokens_pool = Tokens(tokens=[token1, token2])

        order = Order(account_id=7, nonce=18, token_sell=token1, token_buy=token2,
                      ratio=Fraction(1, 4), amount=1000000,
                      recipient='0x823b6a996cea19e0c41e250b20e2e804ea72ccdf',
                      valid_from=0, valid_until=4294967295)
        order.signature = zksync_signer.sign_tx(order)
        order.eth_signature = ethereum_signer.sign_tx(order)
        zksync_validator = EncodedTxValidator(self.library)
        serialized_order = json.dumps(order.dict(), indent=4)

        deserialized_order = Order.from_json(json.loads(serialized_order), tokens_pool)
        ret = zksync_validator.is_valid_signature(deserialized_order)
        self.assertTrue(ret)
        ret = deserialized_order.is_valid_eth_signature(ethereum_signer.address())
        self.assertTrue(ret)

    def test_forced_exit_bytes(self):
        tr = ForcedExit(
            target="0x19aa2ed8712072e918632259780e587698ef58df",
            token=Token.eth(),
            fee=1000000, nonce=12, valid_from=0,
            valid_until=4294967295, initiator_account_id=44
        )
        res = "f7010000002c19aa2ed8712072e918632259780e587698ef58df000000007d030000000c000000000000000000000000ffffffff"
        assert tr.encoded_message().hex() == res

    def test_mint_nft_bytes(self):
        tr = MintNFT(
            creator_id=44,
            creator_address="0xedE35562d3555e61120a151B3c8e8e91d83a378a",
            content_hash="0000000000000000000000000000000000000000000000000000000000000123",
            recipient="0x19aa2ed8712072e918632259780e587698ef58df",
            fee=1000000,
            fee_token=Token.eth(),
            nonce=12
        )
        res = "f6010000002cede35562d3555e61120a151b3c8e8e91d83a378a000000000000000000000000000000000000000000000000000000000000012319aa2ed8712072e918632259780e587698ef58df000000007d030000000c"
        assert tr.encoded_message().hex() == res

    def test_withdraw_nft_bytes(self):
        tr = WithdrawNFT(
            account_id=44,
            from_address="0xedE35562d3555e61120a151B3c8e8e91d83a378a",
            to_address="0x19aa2ed8712072e918632259780e587698ef58df",
            fee_token=Token.eth(),
            fee=1000000,
            nonce=12,
            valid_from=0,
            valid_until=4294967295,
            token_id=100000
        )
        res = "f5010000002cede35562d3555e61120a151b3c8e8e91d83a378a19aa2ed8712072e918632259780e587698ef58df000186a0000000007d030000000c000000000000000000000000ffffffff"
        assert tr.encoded_message().hex() == res

    def test_pack(self):
        amounts = [0, 1, 2047, 2047000, 1000000000000000000000000000000000]
        for amount in amounts:
            assert closest_packable_transaction_fee(amount) == amount
            assert closest_packable_amount(amount) == amount

    def test_signature(self):
        account = Account.from_key(PRIVATE_KEY)
        signer = ZkSyncSigner.from_account(account, self.library, ChainId.MAINNET)
        tr = Transfer(from_address="0xedE35562d3555e61120a151B3c8e8e91d83a378a",
                      to_address="0x19aa2ed8712072e918632259780e587698ef58df",
                      token=Token.eth(),
                      amount=1000000000000,
                      fee=1000000,
                      nonce=12,
                      valid_from=0,
                      valid_until=4294967295, account_id=44)
        res = signer.sign_tx(tr)
        assert res.signature == 'b3211c7e15d31d64619e0c7f65fce8c6e45637b5cfc8711478c5a151e6568d875ec7f48e040225fe3cc7f1e7294625cad6d98b4595d007d36ef62122de16ae01'


def check_bytes(a, b):
    res = True
    for i, c in enumerate(a):
        if c != b[i]:
            print(f"Wrong char {i}, {c}, {b[i]}")
            res = False
    assert res



import os
from unittest import TestCase

from zksync_sdk import ZkSyncLibrary


class TestZkSyncLibrary(TestCase):
    def setUp(self):
        self.library = ZkSyncLibrary()

    def test_public_key_hash_from_seed(self):
        seed = b"1" * 32
        key = self.library.private_key_from_seed(seed)
        assert key != seed
        pub_key = self.library.get_public_key(key)
        assert pub_key != key
        pub_key_hash = self.library.get_pubkey_hash(pub_key)
        assert pub_key != pub_key_hash

    def test_sign(self):
        seed = bytes.fromhex("000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f")
        message = bytes.fromhex(
            "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d3e3f")
        key = self.library.private_key_from_seed(seed)
        signature = self.library.sign(key, message)
        pub_key = self.library.get_public_key(key)

        assert key.hex() == "0552a69519d1f3043611126c13489ff4a2a867a1c667b1d9d9031cd27fdcff5a"
        assert signature.hex() == "5462c3083d92b832d540c9068eed0a0450520f6dd2e4ab169de1a46585b394a4292896a2ebca3c0378378963a6bc1710b64c573598e73de3a33d6cec2f5d7403"
        assert pub_key.hex() == "17f3708f5e2b2c39c640def0cf0010fd9dd9219650e389114ea9da47f5874184"
        assert signature != message



import abc
from dataclasses import dataclass
from decimal import Decimal
from fractions import Fraction
from enum import Enum, IntEnum
from typing import List, Optional, Union, Tuple

from pydantic import BaseModel
from zksync_sdk.lib import ZkSyncLibrary
from zksync_sdk.serializers import (int_to_bytes, packed_amount_checked, packed_fee_checked,
                                    serialize_account_id,
                                    serialize_address, serialize_content_hash,
                                    serialize_nonce, serialize_timestamp,
                                    serialize_token_id, serialize_ratio_part)
from zksync_sdk.types.signatures import TxEthSignature, TxSignature
from zksync_sdk.types.auth_types import ChangePubKeyCREATE2, ChangePubKeyEcdsa

DEFAULT_TOKEN_ADDRESS = "0x0000000000000000000000000000000000000000"

TokenLike = Union[str, int]

TRANSACTION_VERSION = 0x01


class EncodedTxType(IntEnum):
    CHANGE_PUB_KEY = 7
    TRANSFER = 5
    WITHDRAW = 3
    FORCED_EXIT = 8
    SWAP = 11
    MINT_NFT = 9
    WITHDRAW_NFT = 10


class RatioType(Enum):
    # ratio that represents the lowest denominations of tokens (wei for ETH, satoshi for BTC etc.)
    wei = 'Wei',
    # ratio that represents tokens themselves
    token = 'Token'


class Token(BaseModel):
    address: str
    id: int
    symbol: str
    decimals: int

    @classmethod
    def eth(cls):
        return cls(id=0,
                   address=DEFAULT_TOKEN_ADDRESS,
                   symbol="ETH",
                   decimals=18)

    def is_eth(self) -> bool:
        return self.symbol == "ETH" and self.address == DEFAULT_TOKEN_ADDRESS

    def decimal_amount(self, amount: int) -> Decimal:
        return Decimal(amount).scaleb(-self.decimals)

    def from_decimal(self, amount: Decimal) -> int:
        return int(amount.scaleb(self.decimals))

    def decimal_str_amount(self, amount: int) -> str:
        d = self.decimal_amount(amount)

        # Creates a string with `self.decimals` numbers after decimal point.
        # Prevents scientific notation (string values like '1E-8').
        # Prevents integral numbers having no decimal point in the string representation.
        d_str = f"{d:.{self.decimals}f}"

        d_str = d_str.rstrip("0")
        if d_str[-1] == ".":
            return d_str + "0"

        if '.' not in d_str:
            return d_str + '.0'

        return d_str


def token_ratio_to_wei_ratio(token_ratio: Fraction, token_sell: Token, token_buy: Token) -> Fraction:
    num = token_sell.from_decimal(Decimal(token_ratio.numerator))
    den = token_buy.from_decimal(Decimal(token_ratio.denominator))
    return Fraction(num, den, _normalize = False)


class Tokens(BaseModel):
    tokens: List[Token]

    def find_by_address(self, address: str) -> Optional[Token]:
        found_token = [token for token in self.tokens if token.address == address]
        if found_token:
            return found_token[0]
        else:
            return None

    def find_by_id(self, token_id: int) -> Optional[Token]:
        found_token = [token for token in self.tokens if token.id == token_id]
        if found_token:
            return found_token[0]
        else:
            return None

    def find_by_symbol(self, symbol: str) -> Optional[Token]:
        found_token = [token for token in self.tokens if token.symbol == symbol]
        if found_token:
            return found_token[0]
        else:
            return None

    def find(self, token: TokenLike) -> Optional[Token]:
        result = None
        if isinstance(token, int):
            result = self.find_by_id(token)

        if isinstance(token, str):
            result = self.find_by_address(address=token)
            if result is None:
                result = self.find_by_symbol(symbol=token)
        return result


class EncodedTx(abc.ABC):
    @abc.abstractmethod
    def encoded_message(self) -> bytes:
        pass

    @abc.abstractmethod
    def human_readable_message(self) -> str:
        pass

    @abc.abstractmethod
    def tx_type(self) -> int:
        pass

    @abc.abstractmethod
    def dict(self):
        pass

    @abc.abstractmethod
    def batch_message_part(self) -> str:
        pass


@dataclass
class ChangePubKey(EncodedTx):
    account_id: int
    account: str
    new_pk_hash: str
    token: Token
    fee: int
    nonce: int
    valid_from: int
    valid_until: int
    eth_auth_data: Union[ChangePubKeyCREATE2, ChangePubKeyEcdsa, None] = None
    eth_signature: Optional[TxEthSignature] = None
    signature: Optional[TxSignature] = None

    def human_readable_message(self) -> str:
        message = f"Set signing key: {self.new_pk_hash.replace('sync:', '').lower()}"
        if self.fee:
            message += f"\nFee: {self.fee} {self.token.symbol}"
        return message

    def batch_message_part(self) -> str:
        message = f"Set signing key: {self.new_pk_hash.replace('sync:', '').lower()}\n"
        if self.fee:
            message += f"Fee: {self.token.decimal_str_amount(self.fee)} {self.token.symbol}\n"
        return message

    def encoded_message(self) -> bytes:
        return b"".join([
            int_to_bytes(0xff - self.tx_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.account_id),
            serialize_address(self.account),
            serialize_address(self.new_pk_hash),
            serialize_token_id(self.token.id),
            packed_fee_checked(self.fee),
            serialize_nonce(self.nonce),
            serialize_timestamp(self.valid_from),
            serialize_timestamp(self.valid_until)
        ])

    def get_eth_tx_bytes(self) -> bytes:
        data = b"".join([
            serialize_address(self.new_pk_hash),
            serialize_nonce(self.nonce),
            serialize_account_id(self.account_id),
        ])
        if self.eth_auth_data is not None:
            data += self.eth_auth_data.encode_message()
        return data

    def get_auth_data(self, signature: str):
        if self.eth_auth_data is None:
            return {"type": "Onchain"}
        elif isinstance(self.eth_auth_data, ChangePubKeyEcdsa):
            return self.eth_auth_data.dict(signature)
        elif isinstance(self.eth_auth_data, ChangePubKeyCREATE2):
            return self.eth_auth_data.dict()

    def dict(self):
        return {
            "type": "ChangePubKey",
            "accountId": self.account_id,
            "account": self.account,
            "newPkHash": self.new_pk_hash,
            "fee_token": self.token.id,
            "fee": str(self.fee),
            "nonce": self.nonce,
            "ethAuthData": self.eth_auth_data,
            "signature": self.signature.dict(),
            "validFrom": self.valid_from,
            "validUntil": self.valid_until,
        }

    @classmethod
    def tx_type(cls):
        return EncodedTxType.CHANGE_PUB_KEY


@dataclass
class Transfer(EncodedTx):
    account_id: int
    from_address: str
    to_address: str
    token: Token
    amount: int
    fee: int
    nonce: int
    valid_from: int
    valid_until: int
    signature: Optional[TxSignature] = None

    def tx_type(self) -> int:
        return EncodedTxType.TRANSFER

    def human_readable_message(self) -> str:
        msg = ""

        if self.amount != 0:
            msg += f"Transfer {self.token.decimal_str_amount(self.amount)} {self.token.symbol} to: {self.to_address.lower()}\n"
        if self.fee != 0:
            msg += f"Fee: {self.token.decimal_str_amount(self.fee)} {self.token.symbol}\n"

        return msg + f"Nonce: {self.nonce}"

    def batch_message_part(self) -> str:
        msg = ""
        if self.amount != 0:
            msg += f"Transfer {self.token.decimal_str_amount(self.amount)} {self.token.symbol} to: {self.to_address.lower()}\n"
        if self.fee != 0:
            msg += f"Fee: {self.token.decimal_str_amount(self.fee)} {self.token.symbol}\n"
        return msg

    def encoded_message(self) -> bytes:
        return b"".join([
            int_to_bytes(0xff - self.tx_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.account_id),
            serialize_address(self.from_address),
            serialize_address(self.to_address),
            serialize_token_id(self.token.id),
            packed_amount_checked(self.amount),
            packed_fee_checked(self.fee),
            serialize_nonce(self.nonce),
            serialize_timestamp(self.valid_from),
            serialize_timestamp(self.valid_until)
        ])

    def dict(self):
        return {
            "type": "Transfer",
            "accountId": self.account_id,
            "from": self.from_address,
            "to": self.to_address,
            "token": self.token.id,
            "fee": str(self.fee),
            "nonce": self.nonce,
            "signature": self.signature.dict(),
            "amount": str(self.amount),
            "validFrom": self.valid_from,
            "validUntil": self.valid_until,
        }


@dataclass
class Withdraw(EncodedTx):
    account_id: int
    from_address: str
    to_address: str
    amount: int
    fee: int
    nonce: int
    valid_from: int
    valid_until: int
    token: Token
    signature: Optional[TxSignature] = None

    def tx_type(self) -> int:
        return EncodedTxType.WITHDRAW

    def human_readable_message(self) -> str:
        msg = ""
        if self.amount != 0:
            msg += f"Withdraw {self.token.decimal_str_amount(self.amount)} {self.token.symbol} to: {self.to_address.lower()}\n"
        if self.fee != 0:
            msg += f"Fee: {self.token.decimal_str_amount(self.fee)} {self.token.symbol}\n"
        return msg + f"Nonce: {self.nonce}"

    def batch_message_part(self) -> str:
        msg = ""
        if self.amount != 0:
            msg += f"Withdraw {self.token.decimal_str_amount(self.amount)} {self.token.symbol} to: {self.to_address.lower()}\n"
        if self.fee != 0:
            msg += f"Fee: {self.token.decimal_str_amount(self.fee)} {self.token.symbol}\n"
        return msg

    def encoded_message(self) -> bytes:
        return b"".join([
            int_to_bytes(0xff - self.tx_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.account_id),
            serialize_address(self.from_address),
            serialize_address(self.to_address),
            serialize_token_id(self.token.id),
            int_to_bytes(self.amount, length=16),
            packed_fee_checked(self.fee),
            serialize_nonce(self.nonce),
            serialize_timestamp(self.valid_from),
            serialize_timestamp(self.valid_until)
        ])

    def dict(self):
        return {
            "type": "Withdraw",
            "accountId": self.account_id,
            "from": self.from_address,
            "to": self.to_address,
            "token": self.token.id,
            "fee": str(self.fee),
            "nonce": self.nonce,
            "signature": self.signature.dict(),
            "amount": str(self.amount),
            "validFrom": self.valid_from,
            "validUntil": self.valid_until,
        }


@dataclass
class ForcedExit(EncodedTx):
    initiator_account_id: int
    target: str
    token: Token
    fee: int
    nonce: int
    valid_from: int
    valid_until: int
    signature: Optional[TxSignature] = None

    def tx_type(self) -> int:
        return EncodedTxType.FORCED_EXIT

    def encoded_message(self) -> bytes:
        return b"".join([
            int_to_bytes(0xff - self.tx_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.initiator_account_id),
            serialize_address(self.target),
            serialize_token_id(self.token.id),
            packed_fee_checked(self.fee),
            serialize_nonce(self.nonce),
            serialize_timestamp(self.valid_from),
            serialize_timestamp(self.valid_until)
        ])

    def human_readable_message(self) -> str:
        message = f"ForcedExit {self.token.symbol} to: {self.target.lower()}\nFee: {self.token.decimal_str_amount(self.fee)} {self.token.symbol}\nNonce: {self.nonce}"
        return message

    def batch_message_part(self) -> str:
        message = f"ForcedExit {self.token.symbol} to: {self.target.lower()}\n" \
                  f"Fee: {self.token.decimal_str_amount(self.fee)} {self.token.symbol}\n"
        return message

    def dict(self):
        return {
            "type": "ForcedExit",
            "initiatorAccountId": self.initiator_account_id,
            "target": self.target,
            "token": self.token.id,
            "fee": str(self.fee),
            "nonce": self.nonce,
            "signature": self.signature.dict(),
            "validFrom": self.valid_from,
            "validUntil": self.valid_until,
        }


@dataclass
class Order(EncodedTx):
    account_id: int
    recipient: str
    nonce: int
    token_sell: Token
    token_buy: Token
    amount: int
    ratio: Fraction
    valid_from: int
    valid_until: int
    signature: Optional[TxSignature] = None
    eth_signature: Optional[TxEthSignature] = None

    @classmethod
    def from_json(cls, json: dict, tokens: Tokens):

        def from_optional(value: Optional[Token]) -> Token:
            if value is None:
                raise ValueError(f"Token None value should not be used")
            return value

        token_sell_id: int = json["tokenSell"]
        token_buy_id: int = json["tokenBuy"]
        token_sell = from_optional(tokens.find_by_id(token_sell_id))
        token_buy = from_optional(tokens.find_by_id(token_buy_id))
        ratio = json["ratio"]

        # INFO: could be None
        signature = json.get("signature")
        if signature is not None:
            signature = TxSignature.from_dict(signature)
        ether_sig = json.get("ethSignature")
        if ether_sig is not None:
            ether_sig = TxEthSignature.from_dict(ether_sig)
        return cls(
            account_id=json["accountId"],
            recipient=json["recipient"],
            nonce=json["nonce"],
            token_sell=token_sell,
            token_buy=token_buy,
            amount=int(json["amount"]),
            ratio=Fraction(int(ratio[0]), int(ratio[1]), _normalize=False),
            valid_from=json["validFrom"],
            valid_until=json["validUntil"],
            signature=signature,
            eth_signature=ether_sig
        )

    def tx_type(self) -> int:
        raise NotImplementedError

    def msg_type(self) -> int:
        return b'o'[0]

    def encoded_message(self) -> bytes:
        return b"".join([
            int_to_bytes(self.msg_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.account_id),
            serialize_address(self.recipient),
            serialize_nonce(self.nonce),
            serialize_token_id(self.token_sell.id),
            serialize_token_id(self.token_buy.id),
            serialize_ratio_part(self.ratio.numerator),
            serialize_ratio_part(self.ratio.denominator),
            packed_amount_checked(self.amount),
            serialize_timestamp(self.valid_from),
            serialize_timestamp(self.valid_until)
        ])

    def batch_message_part(self) -> str:
        pass

    def human_readable_message(self) -> str:
        if self.amount == 0:
            header = f'Limit order for {self.token_sell.symbol} -> {self.token_buy.symbol}'
        else:
            amount = self.token_sell.decimal_str_amount(self.amount)
            header = f'Order for {amount} {self.token_sell.symbol} -> {self.token_buy.symbol}'

        message = '\n'.join([
            header,
            f'Ratio: {self.ratio.numerator}:{self.ratio.denominator}',
            f'Address: {self.recipient.lower()}',
            f'Nonce: {self.nonce}'
        ])
        return message

    def dict(self):
        return {
            "accountId": self.account_id,
            "recipient": self.recipient,
            "nonce": self.nonce,
            "tokenSell": self.token_sell.id,
            "tokenBuy": self.token_buy.id,
            "amount": str(self.amount),
            "ratio": (str(self.ratio.numerator), str(self.ratio.denominator)),
            "validFrom": self.valid_from,
            "validUntil": self.valid_until,
            "signature": self.signature.dict() if self.signature else None,
            "ethSignature": self.eth_signature.dict() if self.eth_signature else None,
        }

    def is_valid_eth_signature(self, signer_address: str) -> bool:
        address = self._recover_signer_address()
        return signer_address == address

    def _recover_signer_address(self) -> str:
        """
        INFO: particular case implementation with dependency from Web3 interface
              if it's needed to generelize for all Obejct type(Transfer, Swap etc) move to etherium_signer module
              with interface & implemnetation for Web3 as Validator class( Visitor pattern )
        """
        from web3.auto import w3
        from eth_account.messages import encode_defunct

        msg = self.human_readable_message().encode()
        encoded_message = encode_defunct(msg)

        def get_sig(opt_value: Optional[TxEthSignature]) -> TxEthSignature:
            if opt_value is None:
                raise ValueError()
            return opt_value

        # INFO: remove prefix 0x
        eth_sig = get_sig(self.eth_signature)
        sig = bytes.fromhex(eth_sig.signature[2:])
        return w3.eth.account.recover_message(encoded_message, signature=sig)


@dataclass
class Swap(EncodedTx):
    submitter_id: int
    submitter_address: str
    amounts: Tuple[int, int]
    orders: Tuple[Order, Order]
    fee_token: Token
    fee: int
    nonce: int
    signature: Optional[TxSignature] = None

    def tx_type(self) -> int:
        return EncodedTxType.SWAP

    def human_readable_message(self) -> str:
        if self.fee != 0:
            message = f'Swap fee: {self.fee_token.decimal_str_amount(self.fee)} {self.fee_token.symbol}\n'
        else:
            message = ''
        message += f'Nonce: {self.nonce}'
        return message

    def batch_message_part(self) -> str:
        if self.fee != 0:
            message = f'Swap fee: {self.fee_token.decimal_str_amount(self.fee)} {self.fee_token.symbol}\n'
        else:
            message = ''
        return message

    def encoded_message(self) -> bytes:
        order_bytes = b''.join([
            self.orders[0].encoded_message(),
            self.orders[1].encoded_message(),
        ])
        return b"".join([
            int_to_bytes(0xff - self.tx_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.submitter_id),
            serialize_address(self.submitter_address),
            serialize_nonce(self.nonce),
            ZkSyncLibrary().hash_orders(order_bytes),
            serialize_token_id(self.fee_token.id),
            packed_fee_checked(self.fee),
            packed_amount_checked(self.amounts[0]),
            packed_amount_checked(self.amounts[1]),
        ])

    def dict(self):
        return {
            "type": "Swap",
            "submitterId": self.submitter_id,
            "submitterAddress": self.submitter_address,
            "feeToken": self.fee_token.id,
            "fee": str(self.fee),
            "nonce": self.nonce,
            "signature": self.signature.dict() if self.signature else None,
            "amounts": (str(self.amounts[0]), str(self.amounts[1])),
            "orders": (self.orders[0].dict(), self.orders[1].dict())
        }


@dataclass
class MintNFT(EncodedTx):
    creator_id: int
    creator_address: str
    content_hash: str
    recipient: str
    fee: int
    fee_token: Token
    nonce: int
    signature: Optional[TxSignature] = None

    def tx_type(self) -> int:
        return EncodedTxType.MINT_NFT

    def encoded_message(self) -> bytes:
        return b"".join([
            int_to_bytes(0xff - self.tx_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.creator_id),
            serialize_address(self.creator_address),
            serialize_content_hash(self.content_hash),
            serialize_address(self.recipient),
            serialize_token_id(self.fee_token.id),
            packed_fee_checked(self.fee),
            serialize_nonce(self.nonce),
        ])

    def human_readable_message(self) -> str:
        message = f"MintNFT {self.content_hash} for: {self.recipient.lower()}\n" \
                  f"Fee: {self.fee_token.decimal_str_amount(self.fee)} {self.fee_token.symbol}\nNonce: {self.nonce}"
        return message

    def batch_message_part(self) -> str:
        message = f"MintNFT {self.content_hash} for: {self.recipient.lower()}\n" \
                  f"Fee: {self.fee_token.decimal_str_amount(self.fee)} {self.fee_token.symbol}\n"
        return message

    def dict(self):
        return {
            "type": "MintNFT",
            "creatorId": self.creator_id,
            "creatorAddress": self.creator_address,
            "contentHash": self.content_hash,
            "recipient": self.recipient,
            "feeToken": self.fee_token.id,
            "fee": str(self.fee),
            "nonce": self.nonce,
            "signature": self.signature.dict(),
        }


@dataclass
class WithdrawNFT(EncodedTx):
    account_id: int
    from_address: str
    to_address: str
    fee_token: Token
    fee: int
    nonce: int
    valid_from: int
    valid_until: int
    token_id: int
    signature: Optional[TxSignature] = None

    def tx_type(self) -> int:
        return EncodedTxType.WITHDRAW_NFT

    def encoded_message(self) -> bytes:
        return b"".join([
            int_to_bytes(0xff - self.tx_type(), 1),
            int_to_bytes(TRANSACTION_VERSION, 1),
            serialize_account_id(self.account_id),
            serialize_address(self.from_address),
            serialize_address(self.to_address),
            serialize_token_id(self.token_id),
            serialize_token_id(self.fee_token.id),
            packed_fee_checked(self.fee),
            serialize_nonce(self.nonce),
            serialize_timestamp(self.valid_from),
            serialize_timestamp(self.valid_until)
        ])

    def human_readable_message(self) -> str:
        message = f"WithdrawNFT {self.token_id} to: {self.to_address.lower()}\nFee: {self.fee_token.decimal_str_amount(self.fee)} {self.fee_token.symbol}\nNonce: {self.nonce}"
        return message

    def batch_message_part(self) -> str:
        message = f"WithdrawNFT {self.token_id} to: {self.to_address.lower()}\n" \
                  f"Fee: {self.fee_token.decimal_str_amount(self.fee)} {self.fee_token.symbol}\n"
        return message

    def dict(self):
        return {
            "type": "WithdrawNFT",
            "accountId": self.account_id,
            "from": self.from_address,
            "to": self.to_address,
            "feeToken": self.fee_token.id,
            "fee": str(self.fee),
            "nonce": self.nonce,
            "validFrom": self.valid_from,
            "validUntil": self.valid_until,
            "token": self.token_id,
            "signature": self.signature.dict(),
        }


class EncodedTxValidator:
    def __init__(self, library: ZkSyncLibrary):
        self.library = library

    def is_valid_signature(self, tx):
        zk_sync_signature: TxSignature = tx.signature
        if zk_sync_signature is None:
            return False
        bytes_signature = bytes.fromhex(zk_sync_signature.signature)
        pubkey = bytes.fromhex(zk_sync_signature.public_key)
        return self.library.is_valid_signature(tx.encoded_message(), pubkey, bytes_signature)


@dataclass
class TransactionWithSignature:
    tx: EncodedTx
    signature: TxEthSignature

    def dict(self):
        return {
            'tx': self.tx.dict(),
            'signature': self.signature.dict(),
        }


@dataclass()
class TransactionWithOptionalSignature:
    tx: EncodedTx
    signature: Union[None, TxEthSignature, List[TxSignature]] = None

    def dict(self):
        if self.signature is None:
            null_value = None
            return {
                'signature': null_value,
                'tx': self.tx.dict()
            }
        else:
            if isinstance(self.signature, list):
                null_value = None
                value = []
                for sig in self.signature:
                    if sig is None:
                        value.append(null_value)
                    else:
                        value.append(sig.dict())
                return {
                    'signature': value,
                    'tx': self.tx.dict()
                }
            else:
                return {
                    'signature': self.signature.dict(),
                    'tx': self.tx.dict()
                }



from enum import Enum

__all__ = ['FeeTxType']


class FeeTxType(Enum):
    withdraw = "Withdraw"
    transfer = "Transfer"
    swap = "Swap"
    fast_withdraw = "FastWithdraw"
    change_pub_key_onchain = {"ChangePubKey": "Onchain"}
    change_pub_key_ecdsa = {"ChangePubKey": "ECDSA"}
    change_pub_key_create2 = {"ChangePubKey": "CREATE2"}
    mint_nft = "MintNFT"
    withdraw_nft = "WithdrawNFT"



from eth_account.signers.base import BaseAccount
from web3 import Web3

from zksync_sdk.contract_utils import erc20_abi, zksync_abi

MAX_ERC20_APPROVE_AMOUNT = 115792089237316195423570985008687907853269984665640564039457584007913129639935  # 2^256 - 1
ERC20_APPROVE_THRESHOLD = 57896044618658097711785492504343953926634992332820282019728792003956564819968  # 2^255


class Contract:
    def __init__(self, contract_address: str, web3: Web3, account: BaseAccount, abi):
        self.contract_address = contract_address
        self.web3 = web3
        self.contract = self.web3.eth.contract(self.contract_address, abi=abi)  # type: ignore[call-overload]
        self.account = account

    def _call_method(self, method_name, *args, amount=None, **kwargs):
        params = {}
        if amount is not None:
            params['value'] = amount
        params['from'] = self.account.address
        transaction = getattr(self.contract.functions, method_name)(
            *args,
            **kwargs
        ).buildTransaction(params)

        transaction.update({'nonce': self.web3.eth.get_transaction_count(self.account.address)})
        signed_tx = self.account.sign_transaction(transaction)
        txn_hash = self.web3.eth.send_raw_transaction(signed_tx.rawTransaction)
        txn_receipt = self.web3.eth.waitForTransactionReceipt(txn_hash)
        return txn_receipt


class ZkSync(Contract):
    def __init__(self, web3: Web3, zksync_contract_address: str, account: BaseAccount):
        super().__init__(zksync_contract_address, web3, account, zksync_abi())

    def deposit_eth(self, address: str, amount: int):
        return self._call_method("depositETH", address, amount=amount)

    def deposit_erc20(self, token_address: str, address: str, amount: int):
        return self._call_method("depositERC20", token_address, amount, address)

    def full_exit(self, account_id: int, token_address: str, ):
        return self._call_method("requestFullExit", account_id, token_address)

    def full_exit_nft(self, account_id: int, token_id: int):
        return self._call_method("requestFullExitNFT", account_id, token_id)

    def set_auth_pub_key_hash(self, pub_key_hash: bytes, nonce: int):
        return self._call_method("setAuthPubkeyHash", pub_key_hash, nonce)

    def auth_facts(self, sender_address: str, nonce: int) -> bytes:
        return self.contract.caller.authFacts(sender_address, nonce)


class ERC20Contract(Contract):
    def __init__(self, web3: Web3, zksync_address: str, contract_address: str,
                 account: BaseAccount):
        self.zksync_address = zksync_address
        super().__init__(contract_address, web3, account, erc20_abi())

    def approve_deposit(self, max_erc20_approve_amount=MAX_ERC20_APPROVE_AMOUNT):
        return self._call_method('approve', self.zksync_address, max_erc20_approve_amount)

    def is_deposit_approved(self, erc20_approve_threshold=ERC20_APPROVE_THRESHOLD):
        allowance = self.contract.functions.allowance(self.account.address,
                                                      self.zksync_address).call()

        return allowance >= erc20_approve_threshold



import time
from decimal import Decimal
from fractions import Fraction
from typing import List, Optional, Tuple, Union

from zksync_sdk.ethereum_provider import EthereumProvider
from zksync_sdk.ethereum_signer import EthereumSignerInterface
from zksync_sdk.types import (ChangePubKey, ChangePubKeyCREATE2, ChangePubKeyEcdsa,
                              ChangePubKeyTypes, EncodedTx, ForcedExit, Token, TokenLike,
                              Tokens, TransactionWithSignature, Transfer, TxEthSignature,
                              Withdraw, MintNFT, WithdrawNFT, NFT, Order, Swap, RatioType,
                              token_ratio_to_wei_ratio, get_toggle_message, get_toggle_message_with_pub, Toggle2FA)
from zksync_sdk.zksync_provider import FeeTxType, ZkSyncProviderInterface
from zksync_sdk.zksync_signer import ZkSyncSigner
from zksync_sdk.zksync_provider.transaction import Transaction

DEFAULT_VALID_FROM = 0
DEFAULT_VALID_UNTIL = 2 ** 32 - 1


class WalletError(Exception):
    pass


class TokenNotFoundError(WalletError):
    pass


class AmountsMissing(WalletError):
    pass


class Wallet:
    def __init__(self, ethereum_provider: EthereumProvider, zk_signer: ZkSyncSigner,
                 eth_signer: EthereumSignerInterface, provider: ZkSyncProviderInterface):
        self.ethereum_provider = ethereum_provider
        self.zk_signer = zk_signer
        self.eth_signer = eth_signer
        self.zk_provider = provider
        self.account_id = None
        self.tokens = Tokens(tokens=[])

    async def get_account_id(self):
        if self.account_id is None:
            state = await self.zk_provider.get_state(self.address())
            if isinstance(state.id, int):
                self.account_id = state.id
        return self.account_id

    async def send_signed_transaction(self, tx: EncodedTx,
                                      eth_signature: Union[Optional[TxEthSignature], List[Optional[TxEthSignature]]],
                                      fast_processing: bool = False) -> Transaction:
        return await self.zk_provider.submit_tx(tx, eth_signature, fast_processing)

    async def send_txs_batch(self, transactions: List[TransactionWithSignature],
                             signatures: Optional[
                                 Union[List[TxEthSignature], TxEthSignature]
                             ] = None) -> List[Transaction]:
        return await self.zk_provider.submit_txs_batch(transactions, signatures)

    async def set_signing_key(self, fee_token: TokenLike, *,
                              eth_auth_data: Union[ChangePubKeyCREATE2, ChangePubKeyEcdsa, None] = None,
                              fee: Optional[Decimal] = None, nonce: Optional[int] = None,
                              valid_from=DEFAULT_VALID_FROM, valid_until=DEFAULT_VALID_UNTIL):
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        fee_token_obj = await self.resolve_token(fee_token)
        if isinstance(eth_auth_data, ChangePubKeyEcdsa):
            eth_auth_type = ChangePubKeyTypes.ecdsa
        elif isinstance(eth_auth_data, ChangePubKeyCREATE2):
            eth_auth_type = ChangePubKeyTypes.create2
        else:
            eth_auth_type = ChangePubKeyTypes.onchain

        if fee is None:
            if eth_auth_type == ChangePubKeyTypes.ecdsa:
                fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.change_pub_key_ecdsa,
                                                                 self.address(),
                                                                 fee_token_obj.id)
            elif eth_auth_type == ChangePubKeyTypes.onchain:
                fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.change_pub_key_onchain,
                                                                 self.address(),
                                                                 fee_token_obj.id)
            else:
                assert eth_auth_type == ChangePubKeyTypes.create2, "invalid eth_auth_type"
                fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.change_pub_key_create2,
                                                                 self.address(),
                                                                 fee_token_obj.id)
            fee_int = fee_obj.total_fee
        else:
            fee_int = fee_token_obj.from_decimal(fee)

        change_pub_key, eth_signature = await self.build_change_pub_key(fee_token_obj,
                                                                        eth_auth_data, fee_int,
                                                                        nonce,
                                                                        valid_from,
                                                                        valid_until)

        return await self.send_signed_transaction(change_pub_key, eth_signature)

    # This function takes as a parameter the integer fee of 
    # lowest token denominations (wei, satoshi, etc.)
    async def build_change_pub_key(
            self,
            fee_token: Token,
            eth_auth_data: Union[ChangePubKeyCREATE2, ChangePubKeyEcdsa, None],
            fee: int,
            nonce: Optional[int] = None,
            valid_from=DEFAULT_VALID_FROM,
            valid_until=DEFAULT_VALID_UNTIL):
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        account_id = await self.get_account_id()

        new_pubkey_hash = self.zk_signer.pubkey_hash_str()
        change_pub_key = ChangePubKey(
            account=self.address(),
            account_id=account_id,
            new_pk_hash=new_pubkey_hash,
            token=fee_token,
            fee=fee,
            nonce=nonce,
            valid_until=valid_until,
            valid_from=valid_from,
            eth_auth_data=eth_auth_data
        )

        eth_signature = self.eth_signer.sign(change_pub_key.get_eth_tx_bytes())
        eth_auth_data = change_pub_key.get_auth_data(eth_signature.signature)

        change_pub_key.eth_auth_data = eth_auth_data
        zk_signature = self.zk_signer.sign_tx(change_pub_key)
        change_pub_key.signature = zk_signature

        return change_pub_key, eth_signature

    async def forced_exit(self, target: str, token: TokenLike, fee: Optional[Decimal] = None,
                          valid_from=DEFAULT_VALID_FROM, valid_until=DEFAULT_VALID_UNTIL) -> Transaction:
        nonce = await self.zk_provider.get_account_nonce(self.address())
        token_obj = await self.resolve_token(token)
        if fee is None:
            fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.withdraw, target, token_obj.id)
            fee_int = fee_obj.total_fee
        else:
            fee_int = token_obj.from_decimal(fee)

        transfer, eth_signature = await self.build_forced_exit(target, token_obj, fee_int, nonce,
                                                               valid_from, valid_until)

        return await self.send_signed_transaction(transfer, eth_signature)

    # This function takes as a parameter the integer fee of 
    # lowest token denominations (wei, satoshi, etc.)
    async def build_forced_exit(
            self,
            target: str,
            token: Token,
            fee: int,
            nonce: Optional[int] = None,
            valid_from=DEFAULT_VALID_FROM,
            valid_until=DEFAULT_VALID_UNTIL) -> Tuple[ForcedExit, TxEthSignature]:
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        account_id = await self.get_account_id()

        forced_exit = ForcedExit(initiator_account_id=account_id,
                                 target=target,
                                 fee=fee,
                                 nonce=nonce,
                                 valid_from=valid_from,
                                 valid_until=valid_until,
                                 token=token)
        eth_signature = self.eth_signer.sign_tx(forced_exit)
        zk_signature = self.zk_signer.sign_tx(forced_exit)
        forced_exit.signature = zk_signature

        return forced_exit, eth_signature

    async def mint_nft(self, content_hash: str, recipient: str,token: TokenLike, fee: Optional[Decimal] = None) -> Transaction:
        token_obj = await self.resolve_token(token)

        nonce = await self.zk_provider.get_account_nonce(self.address())
        if fee is None:
            fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.mint_nft, recipient, token_obj.id)
            fee_int = fee_obj.total_fee
        else:
            fee_int = token_obj.from_decimal(fee)

        mint_nft, eth_signature = await self.build_mint_nft(content_hash, recipient, token_obj, fee_int, nonce)
        return await self.send_signed_transaction(mint_nft, eth_signature)

    # This function takes as a parameter the integer fee of 
    # lowest token denominations (wei, satoshi, etc.)
    async def build_mint_nft(
            self,
            content_hash: str,
            recipient: str,
            token: Token,
            fee: int,
            nonce: Optional[int] = None
    ) -> Tuple[MintNFT, TxEthSignature]:
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        account_id = await self.get_account_id()

        mint_nft = MintNFT(creator_id=account_id,
                           creator_address=self.address(),
                           content_hash=content_hash,
                           recipient=recipient,
                           fee=fee,
                           fee_token=token,
                           nonce=nonce)
        eth_signature = self.eth_signer.sign_tx(mint_nft)
        zk_signature = self.zk_signer.sign_tx(mint_nft)
        mint_nft.signature = zk_signature

        return mint_nft, eth_signature

    async def withdraw_nft(
            self,
            to_address: str,
            nft_token: NFT,
            fee_token: TokenLike,
            fee: Optional[Decimal] = None,
            valid_from=DEFAULT_VALID_FROM,
            valid_until=DEFAULT_VALID_UNTIL
    ) -> Transaction:
        nonce = await self.zk_provider.get_account_nonce(self.address())
        fee_token_obj = await self.resolve_token(fee_token)

        if fee is None:
            fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.withdraw_nft, to_address, fee_token_obj.id)
            fee_int = fee_obj.total_fee
        else:
            fee_int = fee_token_obj.from_decimal(fee)
        withdraw_nft, eth_signature = await self.build_withdraw_nft(to_address, nft_token, fee_token_obj, fee_int,
                                                                    nonce, valid_from, valid_until)
        return await self.send_signed_transaction(withdraw_nft, eth_signature)

    # This function takes as a parameter the integer fee of 
    # lowest token denominations (wei, satoshi, etc.)
    async def build_withdraw_nft(
            self,
            to_address: str,
            nft_token: NFT,
            fee_token: Token,
            fee: int,
            nonce: Optional[int] = None,
            valid_from=DEFAULT_VALID_FROM,
            valid_until=DEFAULT_VALID_UNTIL
    ) -> Tuple[WithdrawNFT, TxEthSignature]:
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        account_id = await self.get_account_id()

        withdraw_nft = WithdrawNFT(
            account_id=account_id,
            from_address=self.address(),
            to_address=to_address,
            fee_token=fee_token,
            fee=fee,
            nonce=nonce,
            valid_from=valid_from,
            valid_until=valid_until,
            token_id=nft_token.id)
        eth_signature = self.eth_signer.sign_tx(withdraw_nft)
        zk_signature = self.zk_signer.sign_tx(withdraw_nft)
        withdraw_nft.signature = zk_signature

        return withdraw_nft, eth_signature

    def address(self):
        return self.eth_signer.address()

    async def build_transfer(
            self,
            to: str,
            amount: int,
            token: Token,
            fee: int,
            nonce: Optional[int] = None,
            valid_from: int = DEFAULT_VALID_FROM,
            valid_until: int = DEFAULT_VALID_UNTIL,
    ) -> Tuple[Transfer, TxEthSignature]:
        """
        This function takes as a parameter the integer amount/fee of lowest token denominations (wei, satoshi, etc.)
        """
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        account_id = await self.get_account_id()

        transfer = Transfer(account_id=account_id, from_address=self.address(),
                            to_address=to.lower(),
                            amount=amount, fee=fee,
                            nonce=nonce,
                            valid_from=valid_from,
                            valid_until=valid_until,
                            token=token)
        eth_signature = self.eth_signer.sign_tx(transfer)
        zk_signature = self.zk_signer.sign_tx(transfer)
        transfer.signature = zk_signature
        return transfer, eth_signature

    async def transfer(self, to: str, amount: Decimal, token: TokenLike,
                       fee: Optional[Decimal] = None,
                       valid_from=DEFAULT_VALID_FROM, valid_until=DEFAULT_VALID_UNTIL) -> Transaction:
        nonce = await self.zk_provider.get_account_nonce(self.address())
        token_obj = await self.resolve_token(token)

        if fee is None:
            fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.transfer, to, token_obj.id)
            fee_int = fee_obj.total_fee
        else:
            fee_int = token_obj.from_decimal(fee)

        amount_int = token_obj.from_decimal(amount)

        transfer, eth_signature = await self.build_transfer(to, amount_int, token_obj, fee_int, nonce, valid_from, valid_until)
        return await self.send_signed_transaction(transfer, eth_signature)

    async def transfer_nft(self, to: str, nft: NFT, fee_token: TokenLike,
                           fee: Optional[Decimal] = None,
                           valid_from=DEFAULT_VALID_FROM,
                           valid_until=DEFAULT_VALID_UNTIL
                           ) -> List[Transaction]:
        nonce = await self.zk_provider.get_account_nonce(self.address())
        fee_token_obj = await self.resolve_token(fee_token)

        if fee is None:
            fee_int = await self.zk_provider.get_transactions_batch_fee(
                [FeeTxType.transfer, FeeTxType.transfer],
                [to, self.address()],
                fee_token_obj.symbol
            )
        else:
            fee_int = fee_token_obj.from_decimal(fee)

        nft_tx = await self.build_transfer(to, 1, nft, 0, nonce, valid_from, valid_until)
        fee_tx = await self.build_transfer(self.address(), 0, fee_token_obj, fee_int, nonce + 1, valid_from, valid_until)
        batch = [
            TransactionWithSignature(nft_tx[0], nft_tx[1]),
            TransactionWithSignature(fee_tx[0], fee_tx[1])
        ]
        return await self.send_txs_batch(batch)

    async def get_order(self, token_sell: TokenLike, token_buy: TokenLike,
                        ratio: Fraction, ratio_type: RatioType, amount: Decimal,
                        recipient: Optional[str] = None,
                        nonce: Optional[int] = None,
                        valid_from=DEFAULT_VALID_FROM,
                        valid_until=DEFAULT_VALID_UNTIL) -> Order:
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        token_sell_obj = await self.resolve_token(token_sell)
        token_buy_obj = await self.resolve_token(token_buy)
        recipient = recipient or self.address()

        if ratio_type == RatioType.token:
            ratio = token_ratio_to_wei_ratio(ratio, token_sell_obj, token_buy_obj)

        account_id = await self.get_account_id()

        order = Order(account_id=account_id, recipient=recipient,
                      token_sell=token_sell_obj,
                      token_buy=token_buy_obj,
                      ratio=ratio,
                      amount=token_sell_obj.from_decimal(amount),
                      nonce=nonce,
                      valid_from=valid_from,
                      valid_until=valid_until)

        order.eth_signature = self.eth_signer.sign_tx(order)
        order.signature = self.zk_signer.sign_tx(order)

        return order

    async def get_limit_order(self, token_sell: TokenLike, token_buy: TokenLike,
                              ratio: Fraction, ratio_type: RatioType,
                              recipient: Optional[str] = None,
                              valid_from=DEFAULT_VALID_FROM,
                              valid_until=DEFAULT_VALID_UNTIL):
        return await self.get_order(token_sell, token_buy, ratio, ratio_type, Decimal(0), recipient, valid_from,
                                    valid_until)

    # This function takes as a parameter the integer amounts/fee of 
    # lowest token denominations (wei, satoshi, etc.)
    async def build_swap(self, orders: Tuple[Order, Order], fee_token: Token,
                         amounts: Tuple[int, int], fee: int, nonce: Optional[int] = None):
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        account_id = await self.get_account_id()

        swap = Swap(
            orders=orders, fee_token=fee_token, amounts=amounts, fee=fee, nonce=nonce,
            submitter_id=account_id, submitter_address=self.address()
        )
        eth_signature = self.eth_signer.sign_tx(swap)
        swap.signature = self.zk_signer.sign_tx(swap)
        return swap, eth_signature

    async def swap(self, orders: Tuple[Order, Order], fee_token: TokenLike,
                   amounts: Optional[Tuple[Decimal, Decimal]] = None, fee: Optional[Decimal] = None):
        nonce = await self.zk_provider.get_account_nonce(self.address())

        fee_token_obj = await self.resolve_token(fee_token)
        if fee is None:
            fee_obj = await self.zk_provider.get_transaction_fee(FeeTxType.swap, self.address(), fee_token_obj.id)
            fee_int = fee_obj.total_fee
        else:
            fee_int = fee_token_obj.from_decimal(fee)

        if amounts is None:
            amounts_int = (orders[0].amount, orders[1].amount)
            if amounts_int[0] == 0 or amounts_int[1] == 0:
                raise AmountsMissing("in this case you must specify amounts explicitly")
        else:
            amounts_int = (
                orders[0].token_sell.from_decimal(amounts[0]),
                orders[1].token_sell.from_decimal(amounts[1])
            )

        swap, eth_signature = await self.build_swap(orders, fee_token_obj, amounts_int, fee_int, nonce)
        eth_signatures = [eth_signature, swap.orders[0].eth_signature, swap.orders[1].eth_signature]
        return await self.send_signed_transaction(swap, eth_signatures)

    # This function takes as a parameter the integer amount/fee of 
    # lowest token denominations (wei, satoshi, etc.)
    async def build_withdraw(self, eth_address: str, amount: int, token: Token,
                             fee: int,
                             nonce: Optional[int] = None,
                             valid_from=DEFAULT_VALID_FROM,
                             valid_until=DEFAULT_VALID_UNTIL):
        if nonce is None:
            nonce = await self.zk_provider.get_account_nonce(self.address())
        account_id = await self.get_account_id()

        withdraw = Withdraw(account_id=account_id, from_address=self.address(),
                            to_address=eth_address,
                            amount=amount, fee=fee,
                            nonce=nonce,
                            valid_from=valid_from,
                            valid_until=valid_until,
                            token=token)
        eth_signature = self.eth_signer.sign_tx(withdraw)
        zk_signature = self.zk_signer.sign_tx(withdraw)
        withdraw.signature = zk_signature
        return withdraw, eth_signature

    async def withdraw(self, eth_address: str, amount: Decimal, token: TokenLike,
                       fee: Optional[Decimal] = None, fast: bool = False,
                       valid_from=DEFAULT_VALID_FROM, valid_until=DEFAULT_VALID_UNTIL) -> Transaction:
        nonce = await self.zk_provider.get_account_nonce(self.address())
        token_obj = await self.resolve_token(token)
        if fee is None:
            tx_type = FeeTxType.fast_withdraw if fast else FeeTxType.withdraw
            fee_obj = await self.zk_provider.get_transaction_fee(tx_type, eth_address, token_obj.id)
            fee_int = fee_obj.total_fee
        else:
            fee_int = token_obj.from_decimal(fee)
        amount_int = token_obj.from_decimal(amount)

        withdraw, eth_signature = await self.build_withdraw(eth_address, amount_int, token_obj, fee_int, nonce,
                                                            valid_from, valid_until)
        return await self.send_signed_transaction(withdraw, eth_signature, fast)

    async def get_balance(self, token: TokenLike, type: str):
        account_state = await self.get_account_state()
        token_obj = await self.resolve_token(token)

        if type == "committed":
            token_balance = account_state.committed.balances.get(token_obj.symbol)
        else:
            token_balance = account_state.verified.balances.get(token_obj.symbol)
        if token_balance is None:
            token_balance = 0
        return token_balance

    async def get_account_state(self):
        return await self.zk_provider.get_state(self.address())

    async def is_signing_key_set(self) -> bool:
        account_state = await self.get_account_state()
        signer_pub_key_hash = self.zk_signer.pubkey_hash_str()
        return account_state.id is not None and \
               account_state.committed.pub_key_hash == signer_pub_key_hash

    async def resolve_token(self, token: TokenLike) -> Token:
        resolved_token = self.tokens.find(token)
        if resolved_token is not None:
            return resolved_token
        self.tokens = await self.zk_provider.get_tokens()
        resolved_token = self.tokens.find(token)
        if resolved_token is None:
            raise TokenNotFoundError
        return resolved_token

    async def enable_2fa(self) -> bool:
        mil_seconds = int(time.time() * 1000)
        msg = get_toggle_message(True, mil_seconds)
        eth_sig = self.eth_signer.sign(msg.encode())
        account_id = await self.get_account_id()
        toggle = Toggle2FA(True,
                           account_id,
                           mil_seconds,
                           eth_sig,
                           None
                           )
        return await self.zk_provider.toggle_2fa(toggle)

    async def disable_2fa(self, pub_key_hash: Optional[str]) -> bool:
        mil_seconds = int(time.time() * 1000)
        if pub_key_hash is None:
            msg = get_toggle_message(False, mil_seconds)
        else:
            msg = get_toggle_message_with_pub(False, mil_seconds, pub_key_hash)
        eth_sig = self.eth_signer.sign(msg.encode())
        account_id = await self.get_account_id()
        toggle = Toggle2FA(False,
                           account_id,
                           mil_seconds,
                           eth_sig,
                           pub_key_hash)
        return await self.zk_provider.toggle_2fa(toggle)

    async def disable_2fa_with_pub_key(self):
        pub_key_hash = self.zk_signer.pubkey_hash_str()
        return await self.disable_2fa(pub_key_hash)



from dataclasses import dataclass

from zksync_sdk.types import ChainId


@dataclass
class Network:
    zksync_url: str
    chain_id: ChainId


rinkeby = Network(zksync_url="https://rinkeby-api.zksync.io/jsrpc", chain_id=ChainId.RINKEBY)
ropsten = Network(zksync_url="https://ropsten-api.zksync.io/jsrpc", chain_id=ChainId.ROPSTEN)
mainnet = Network(zksync_url="https://api.zksync.io/jsrpc", chain_id=ChainId.MAINNET)
localhost = Network(zksync_url="http://localhost:3030/jsrpc", chain_id=ChainId.LOCALHOST)



from abc import ABC, abstractmethod
from decimal import Decimal
from typing import List, Optional, Union
from zksync_sdk.transport import JsonRPCTransport
from zksync_sdk.types import (AccountState, ContractAddress, EncodedTx, EthOpInfo, Fee, Token,
                              TokenLike, Tokens, TransactionDetails, TransactionWithSignature,
                              TransactionWithOptionalSignature,
                              TxEthSignature, Toggle2FA, )
from zksync_sdk.zksync_provider.types import FeeTxType
from zksync_sdk.zksync_provider.transaction import Transaction

__all__ = ['ZkSyncProviderInterface']


class ZkSyncProviderInterface(ABC):
    def __init__(self, provider: JsonRPCTransport):
        self.provider = provider

    @abstractmethod
    async def submit_tx(self, tx: EncodedTx, signature: Union[Optional[TxEthSignature], List[Optional[TxEthSignature]]],
                        fast_processing: bool = False) -> Transaction:
        raise NotImplementedError

    @abstractmethod
    async def get_tokens(self) -> Tokens:
        raise NotImplementedError

    @abstractmethod
    async def submit_txs_batch(self, transactions: List[TransactionWithSignature],
                               signatures: Optional[
                                   Union[List[TxEthSignature], TxEthSignature]
                               ] = None) -> List[Transaction]:
        raise NotImplementedError

    @abstractmethod
    async def submit_batch_builder_txs_batch(self, transactions: List[TransactionWithOptionalSignature],
                                             signature: TxEthSignature) -> List[Transaction]:
        raise NotImplementedError

    @abstractmethod
    async def get_contract_address(self) -> ContractAddress:
        raise NotImplementedError

    @abstractmethod
    async def get_state(self, address: str) -> AccountState:
        raise NotImplementedError

    @abstractmethod
    async def get_confirmations_for_eth_op_amount(self) -> int:
        raise NotImplementedError

    @abstractmethod
    async def get_account_nonce(self, address: str) -> int:
        raise NotImplementedError

    @abstractmethod
    async def get_tx_receipt(self, address: str) -> TransactionDetails:
        raise NotImplementedError

    @abstractmethod
    async def get_eth_tx_for_withdrawal(self, withdrawal_hash: str) -> str:
        raise NotImplementedError

    @abstractmethod
    async def get_priority_op_status(self, serial_id: int) -> EthOpInfo:
        raise NotImplementedError

    @abstractmethod
    async def get_transactions_batch_fee(self, tx_types: List[FeeTxType], addresses: List[str],
                                         token_like) -> int:
        raise NotImplementedError

    @abstractmethod
    async def get_transaction_fee(self, tx_type: FeeTxType, address: str,
                                  token_like: TokenLike) -> Fee:
        raise NotImplementedError

    @abstractmethod
    async def get_token_price(self, token: Token) -> Decimal:
        raise NotImplementedError

    @abstractmethod
    async def toggle_2fa(self, toggle2fa: Toggle2FA) -> bool:
        raise NotImplementedError



from eth_account.messages import encode_defunct
from eth_account.signers.base import BaseAccount
from zksync_sdk.ethereum_signer.interface import EthereumSignerInterface
from zksync_sdk.types import EncodedTx, SignatureType, TxEthSignature

__all__ = ['EthereumSignerWeb3']


class EthereumSignerWeb3(EthereumSignerInterface):
    def __init__(self, account: BaseAccount):
        self.account = account

    def sign_tx(self, tx: EncodedTx) -> TxEthSignature:
        message = tx.human_readable_message()
        return self.sign(message.encode())

    def sign(self, message: bytes) -> TxEthSignature:
        signature = self.account.sign_message(encode_defunct(message))
        return TxEthSignature(signature=signature.signature, sig_type=SignatureType.ethereum_signature)

    def address(self) -> str:
        return self.account.address



from abc import ABC, abstractmethod

from zksync_sdk.types import EncodedTx, TxEthSignature

__all__ = ['EthereumSignerInterface']


class EthereumSignerInterface(ABC):

    @abstractmethod
    def sign_tx(self, tx: EncodedTx) -> TxEthSignature:
        raise NotImplementedError

    @abstractmethod
    def sign(self, message: bytes) -> TxEthSignature:
        raise NotImplementedError

    @abstractmethod
    def address(self) -> str:
        raise NotImplementedError



from unittest import TestCase

from web3 import HTTPProvider, Web3, Account

from zksync_sdk.zksync import ZkSync


class TestZkSyncContract(TestCase):
    private_key = "0xa045b52470d306ff78e91b0d2d92f90f7504189125a46b69423dc673fd6b4f3e"

    def setUp(self) -> None:
        self.account = Account.from_key(self.private_key)
        w3 = Web3(HTTPProvider(
            endpoint_uri="https://rinkeby.infura.io/v3/bcf42e619a704151a1b0d95a35cb2e62"))
        self.zksync = ZkSync(account=self.account,
                             web3=w3,
                             zksync_contract_address="0x82F67958A5474e40E1485742d648C0b0686b6e5D")

    def test_deposit_eth(self):
        tx = self.zksync.deposit_eth(self.account.address, 2 * 10 ** 12)
        assert tx['transactionHash']

    def test_full_exit(self):
        tx = self.zksync.full_exit(1, "0x3B00Ef435fA4FcFF5C209a37d1f3dcff37c705aD")
        assert tx['transactionHash']

    def test_auth_facts(self):
        tx = self.zksync.auth_facts(self.account.address, 2)
        assert tx



from decimal import Decimal
from typing import Optional

from web3 import Web3

from zksync_sdk.types import Token
from zksync_sdk.zksync import ERC20Contract, ZkSync

DEFAULT_AUTH_FACTS = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'


class EthereumProvider:
    def __init__(self, web3: Web3, zksync: ZkSync):
        self.web3 = web3
        self.zksync = zksync

    async def approve_deposit(self, token: Token, limit: Decimal):
        contract = ERC20Contract(self.web3, self.zksync.contract_address, token.address,
                                 self.zksync.account)
        return contract.approve_deposit(token.from_decimal(limit))

    async def deposit(self, token: Token, amount: Decimal, address: str):
        if token.is_eth():
            return self.zksync.deposit_eth(address, token.from_decimal(amount))
        else:
            return self.zksync.deposit_erc20(token.address, address, token.from_decimal(amount))

    async def full_exit(self, token: Token, account_id: int):
        return self.zksync.full_exit(account_id, token.address)

    async def full_exit_nft(self, nft: Token, account_id: int):
        return self.zksync.full_exit_nft(account_id, nft.id)

    async def set_auth_pubkey_hash(self, pubkey_hash: bytes, nonce: int):
        return self.zksync.set_auth_pub_key_hash(pubkey_hash, nonce)

    async def is_deposit_approved(self, token: Token, threshold: int) -> bool:
        contract = ERC20Contract(self.web3, self.zksync.contract_address, token.address,
                                 self.zksync.account)
        return contract.is_deposit_approved(threshold)

    async def is_onchain_auth_pubkey_hash_set(self, nonce: int) -> bool:
        auth_facts = self.zksync.auth_facts(self.zksync.account.address, nonce)
        return auth_facts != DEFAULT_AUTH_FACTS



from http.client import OK
from typing import List, Optional

import httpx

from . import JsonRPCTransport, ProviderError, ResponseError
from ..network import Network


class HttpJsonRPCTransport(JsonRPCTransport):
    def __init__(self, network: Network):
        self.network = network

    async def request(self, method: str, params: Optional[List]):
        async with httpx.AsyncClient() as client:
            response = await client.post(self.network.zksync_url, json=self.create_request(method, params))
            if response.status_code == OK:
                result = response.json()
                if "error" in result:
                    data = result["error"]
                    raise ResponseError(data['code'], data['message'])
                else:
                    return result['result']
            else:
                raise ProviderError(response, "Unexpected status code")



from enum import IntEnum

from .responses import *
from .signatures import *
from .transactions import *
from .auth_types import *


class ChainId(IntEnum):
    MAINNET = 1
    RINKEBY = 4
    ROPSTEN = 3
    LOCALHOST = 9
